import { type Macroable, type MCFunctionNode, type PredicateClass } from 'sandstone/core';
import { ContainerCommandNode } from 'sandstone/core/nodes';
import { CommandArguments, FinalCommandOutput } from '../../helpers.js';
import type { ANCHORS, AXES, BLOCKS, COMPARISON_OPERATORS, Coordinates, DIMENSIONS, ENTITY_TYPES, MultipleEntitiesArgument, ObjectiveArgument, Range, Rotation, SingleEntityArgument } from 'sandstone/arguments';
import type { SandstoneCommands } from 'sandstone/commands';
import type { Node } from 'sandstone/core/nodes';
import type { _RawMCFunctionClass } from 'sandstone/core/resources/datapack/mcfunction';
import type { SandstonePack } from 'sandstone/pack';
import type { LiteralUnion } from 'sandstone/utils';
import type { DataPointClass } from 'sandstone/variables/Data';
import type { ObjectiveClass } from 'sandstone/variables/Objective.js';
import type { Score } from 'sandstone/variables/Score.js';
export type SubCommand = [subcommand: string, ...args: unknown[]];
declare class ExecuteCommandPart<MACRO extends boolean> extends CommandArguments<typeof ExecuteCommandNode> {
    protected nestedExecute: (args: SubCommand, executable?: boolean) => ExecuteCommand<MACRO>;
}
export type StoreType = 'byte' | 'short' | 'int' | 'long' | 'float' | 'double';
export type RelationType = 'attacker' | 'controller' | 'leasher' | 'origin' | 'owner' | 'passengers' | 'target' | 'vehicle';
export declare class ExecuteCommandNode extends ContainerCommandNode<SubCommand[]> {
    command: "execute";
    /**
     * Is being used by internals to generate a function call. (eg. Flow visitor)
     */
    isFake: boolean;
    /**
     * By default, the execute is treated as single (execute.run.stuff).
     * This is set to `false` if the execute is used as a function (execute.run(() => { stuff })).
     */
    isSingleExecute: boolean;
    /**
     * In case of a multiple execute, this specifies the base name of the MCFunction.
     */
    givenCallbackName: string | undefined;
    constructor(sandstonePack: SandstonePack, args?: SubCommand[], { isFake, isSingleExecute, givenCallbackName, body, }?: {
        isFake?: boolean | undefined;
        isSingleExecute?: boolean | undefined;
        givenCallbackName?: string | undefined;
        body?: Node[] | undefined;
    });
    get callbackName(): string;
    append: (...nodes: Node[]) => any;
    getValue: () => any;
    createMCFunction: (currentMCFunction: MCFunctionNode | null) => {
        node: ExecuteCommandNode;
        mcFunction?: undefined;
    } | {
        node: ExecuteCommandNode;
        mcFunction: MCFunctionNode;
    };
}
export declare class ExecuteStoreArgsCommand<MACRO extends boolean> extends ExecuteCommandPart<MACRO> {
    /**
     * Saves the final command's return value as tag data within a block entity.
     *
     * Store as a byte, short, int, long, float, or double. If the return value is a decimal, it is rounded first and then multiplied by `scale`.
     *
     * @param targetPos Position of target block
     *
     * @param path Location of the desired tag to hold the value in.
     *
     * @param type Desired data size/type.
     *
     * @param scale Multiplier to apply before storing value. Defaults to 1.
     */
    block: (targetPos: Macroable<Coordinates<MACRO>, MACRO>, path: Macroable<string, MACRO>, type: Macroable<StoreType, MACRO>, scale?: Macroable<number, MACRO>) => ExecuteCommand<MACRO>;
    /**
     * Saves the final command's return value in either a bossbar's current value or its maximum value.
     *
     * @param id ID of the bossbar to target for saving.
     *
     * @param type Whether to overwrite the bossbar's current value or its max value.
     */
    bossbar: (id: Macroable<string, MACRO>, type: Macroable<'max' | 'value', MACRO>) => ExecuteCommand<MACRO>;
    /**
     * Save the final command's return value in one of an entity's data tags.
     *
     * Store as a byte, short, int, long, float, or double. If the return value is a decimal, it is rounded first and then multiplied by `scale`.
     *
     * Like the `/data` command, `/execute store <arguments>` cannot modify player NBT.
     *
     * @param target A single entity to store under.
     *
     * @param path Location of the desired tag to hold the value in.
     *
     * @param type Desired data size/type.
     *
     * @param scale Multiplier to apply before storing value. Defaults to 1.
     */
    entity: (target: Macroable<SingleEntityArgument<MACRO>, MACRO>, path: Macroable<string, MACRO>, type: Macroable<StoreType, MACRO>, scale?: Macroable<number, MACRO>) => ExecuteCommand<MACRO>;
    /**
     * Overrides the given score with the final command's return value.
     *
     * @param targets Specifies score holder(s) whose score is to be overridden.
     *
     * @param objective A scoreboard objective.
     *
     * @param playerScore The player's score to override.
     */
    score(...args: [targets: Macroable<MultipleEntitiesArgument<MACRO>, MACRO>, objective: Macroable<ObjectiveArgument, MACRO>] | [playerScore: Macroable<Score, MACRO>]): ExecuteCommand<MACRO>;
    /**
     * Uses the `path` within storage `target` to store the return value in.
     *
     * Store as a byte, short, int, long, float, or double. If the return value is a decimal, it is rounded first and then multiplied by `scale`.
     *
     * If the storage does not yet exist, it gets created.
     *
     * @param target Target storage container, as a namespaced ID.
     *
     * @param path Location of the desired tag to hold the value in.
     *
     * @param type Desired data size/type.
     *
     * @param scale Multiplier to apply before storing value. Defaults to 1.
     */
    storage: (target: Macroable<string, MACRO>, path: Macroable<string, MACRO>, type: Macroable<StoreType, MACRO>, scale?: Macroable<number, MACRO>) => ExecuteCommand<MACRO>;
}
export declare class ExecuteStoreCommand<MACRO extends boolean> extends ExecuteCommandPart<MACRO> {
    /** Store the final command's result value. */
    get result(): ExecuteStoreArgsCommand<MACRO>;
    /** Store the final command's success value. */
    get success(): ExecuteStoreArgsCommand<MACRO>;
}
export declare class ExecuteDataArgsCommand<MACRO extends boolean> extends ExecuteCommandPart<MACRO> {
    /**
     * Checks whether the targeted block has any data for a given tag
     * @param pos Position of the block to be tested.
     * @param path Data tag to check for.
     */
    block: (pos: Macroable<Coordinates<MACRO>, MACRO>, path: Macroable<string, MACRO>) => ExecuteCommand<MACRO>;
    /**
     * Checks whether the targeted entity has any data for a given tag
     * @param target One single entity to be tested.
     * @param path Data tag to check for.
     */
    entity: (target: Macroable<SingleEntityArgument<MACRO>, MACRO>, path: Macroable<string, MACRO>) => ExecuteCommand<MACRO>;
    /**
     * Checks whether the targeted storage has any data for a given tag
     * @param source The storage to check in.
     * @param path Data tag to check for.
     */
    storage: (source: Macroable<string, MACRO>, path: Macroable<string, MACRO>) => ExecuteCommand<MACRO>;
}
export declare class ExecuteIfUnlessCommand<MACRO extends boolean> extends ExecuteCommandPart<MACRO> {
    /**
     * Compares the block at a given position to a given block.
     *
     * @param pos Position of a target block to test.
     *
     * @param block A block to test against.
     */
    block: (pos: Macroable<Coordinates<MACRO>, MACRO>, block: Macroable<LiteralUnion<BLOCKS>, MACRO>) => ExecuteCommand<MACRO>;
    /**
     * Checks whether one or more entities exist.
     *
     * @param targets The target entities to check.
     */
    entity: (targets: Macroable<MultipleEntitiesArgument<MACRO>, MACRO>) => ExecuteCommand<MACRO>;
    /**
     * Checks whether the `predicate` evaluates to a positive result.
     *
     * @param predicate The predicate to test.
     */
    predicate(predicate: Macroable<string | PredicateClass, MACRO>): ExecuteCommand<MACRO>;
    /**
     * Check a score against either another score or a given range.
     */
    score(firstTarget: Macroable<SingleEntityArgument<MACRO>, MACRO>, firstObjective: Macroable<string | ObjectiveClass, MACRO>, comparison: 'matches', value: Macroable<Range<MACRO>, MACRO>): ExecuteCommand<MACRO>;
    score(firstTarget: Macroable<SingleEntityArgument<MACRO>, MACRO>, firstObjective: Macroable<string | ObjectiveClass, MACRO>, comparison: Macroable<COMPARISON_OPERATORS, MACRO>, otherTarget: Macroable<SingleEntityArgument<MACRO>, MACRO>, otherObjective: Macroable<string | ObjectiveClass, true>): ExecuteCommand<MACRO>;
    score(firstScore: Score, comparison: 'matches', value: Macroable<Range<MACRO>, MACRO>): ExecuteCommand<MACRO>;
    score(firstScore: Score, comparison: Macroable<COMPARISON_OPERATORS, MACRO>, otherScore: Macroable<Score, MACRO>): ExecuteCommand<MACRO>;
    /**
     * Checks whether the current dimension matches the given one.
     * @param dimension Dimension to check against.
     */
    dimension: (dimension: Macroable<LiteralUnion<DIMENSIONS>, MACRO>) => ExecuteCommand<MACRO>;
    /**
     * Checks whether the targeted block is in a chunk that is fully loaded (entity-processing chunk).
     * @param pos Position of target block to test.
     */
    loaded: (pos: Macroable<Coordinates<MACRO>, MACRO>) => ExecuteCommand<MACRO>;
    /** Checks whether the data point exists or the targeted block, entity or storage has any data for a given tag. */
    data(dataPoint: DataPointClass): void;
    data(): ExecuteDataArgsCommand<MACRO>;
    function(func: Macroable<_RawMCFunctionClass<[], []> | (() => any) | string, MACRO>): ExecuteCommand<MACRO>;
}
export declare class ExecuteFacingEntityCommand<MACRO extends boolean> extends ExecuteCommandPart<MACRO> {
    /**
     * Sets the command rotation to face a given point, as viewed from its anchor (either the eyes or the feet).
     *
     * @param targets The target(s) to rotate towards. (if you target multiple entities subcommands will run for each entity)
     *
     * @param anchor Whether to point at the target's eyes or feet.
     */
    entity: (targets: Macroable<MultipleEntitiesArgument<MACRO>, MACRO>, anchor?: Macroable<ANCHORS, MACRO>) => ExecuteCommand<MACRO>;
}
export declare class ExecutePositionedCommand<MACRO extends boolean> extends ExecuteCommandPart<MACRO> {
    /**
     * Sets the command position, without changing rotation or dimension, by matching an entity's position.
     *
     * @param targets Target entity/entities to match position with.
     */
    as: (targets: Macroable<MultipleEntitiesArgument<MACRO>, MACRO>) => ExecuteCommand<MACRO>;
    /**
     * Sets the command position matching the height map (highest position in a column of blocks according to criteria) for the current position.
     *
     * - `world_surface` Any non-air block.
     * - `motion_blocking` Any motion blocking material (eg. ignores flowers and grass).
     * - `motion_blocking_no_leaves` Any non-leaf motion blocking material.
     * - `ocean_floor` Any non-fluid motion blocking material.
     */
    over: (heightMap: Macroable<'world_surface' | 'motion_blocking' | 'motion_blocking_no_leaves' | 'ocean_floor', MACRO>) => ExecuteCommand<MACRO>;
}
export declare class ExecuteRotatedAsCommand<MACRO extends boolean> extends ExecuteCommandPart<MACRO> {
    /**
     * Sets the command rotation, by matching an entity's rotation.
     *
     * @param targets Target entity/entities to match rotation with.
     */
    as: (targets: Macroable<MultipleEntitiesArgument<MACRO>, MACRO>) => ExecuteCommand<MACRO>;
}
export declare class ExecuteCommand<MACRO extends boolean> extends ExecuteCommandPart<MACRO> {
    protected NodeType: typeof ExecuteCommandNode;
    /**
     * Updates the command's position, aligning to its current block position (an integer). Only applies along specified axes.
     * This is akin to flooring the coordinates – i.e. rounding them downwards. It updates the meaning of `~ ~ ~` and `^ ^ ^`.
     *
     * @param axes Any non-repeating combination of the characters 'x', 'y', and 'z'.
     */
    align: (axes: Macroable<AXES, MACRO>) => ExecuteCommand<MACRO>;
    /**
     * Stores the distance from the feet to the eyes of the entity that is executing the command in the anchor, which is part of the command context.
     * Effectively re-centers `^ ^ ^` on either the eyes or feet, also changing the angle the `facing` sub-command works off of.
     *
     * @param anchor Whether to anchor the executed command to eye level or feet level
     */
    anchored: (anchor: Macroable<ANCHORS, MACRO>) => ExecuteCommand<MACRO>;
    /**
     * Sets the command sender to target entity, without changing position, rotation, dimension, or anchor
     *
     * @param targets Target entity/entities to become the new sender.
     */
    as: (targets: Macroable<MultipleEntitiesArgument<MACRO>, MACRO>) => ExecuteCommand<MACRO>;
    /**
     * Sets the command position, rotation, and dimension to match those of an entity/entities; does not change sender
     * @param targets Target entity/entities to match position, rotation, and dimension with
     */
    at: (targets: Macroable<MultipleEntitiesArgument<MACRO>, MACRO>) => ExecuteCommand<MACRO>;
    /**
     * Sets the command rotation to face a given point, as viewed from its anchor (either the eyes or the feet).
     *
     * @param pos Coordinate to rotate towards.
     */
    facing(pos: Macroable<Coordinates<MACRO>, MACRO>): ExecuteCommand<MACRO>;
    /**
     * Sets the command dimension.
     *
     * @param dimension Name of the new execution dimension.
     */
    in: (dimension: Macroable<LiteralUnion<DIMENSIONS>, MACRO>) => ExecuteCommand<MACRO>;
    /**
     * Sets the command executor to a related entity.
     * @param relation Relation with the current executor.
     */
    on: (relation: Macroable<RelationType, MACRO>) => ExecuteCommand<MACRO>;
    /**
     * Sets the command position, without changing rotation or dimension.
     *
     * @param pos The new position.
     */
    positioned(pos: Macroable<Coordinates<MACRO>, MACRO>): ExecuteCommand<MACRO>;
    /**
     * Sets the command rotation.
     *
     * @param rotation The desired rotation.
     *
     * First value is measured clockwise in degrees from due south (the +Z Axis), ranging [–180 to +180)
     * Second value is measured as declination from the horizon in degrees, ranging [–90 to +90] (straight up to straight down)
     *
     * Relative values can be used to specify a rotation relative to the current execution rotation.
     */
    rotated(rotation: Macroable<Rotation<MACRO>, MACRO>): ExecuteCommand<MACRO>;
    /**
     * Summons an entity at the command position and sets the command executor to said newly summoned entity.
     * @param entityType Entity to summon.
     */
    summon(entityType: Macroable<LiteralUnion<ENTITY_TYPES>, MACRO>): ExecuteCommand<MACRO>;
    get if(): ExecuteIfUnlessCommand<MACRO>;
    get unless(): ExecuteIfUnlessCommand<MACRO>;
    get store(): ExecuteStoreCommand<MACRO>;
    get run(): SandstoneCommands<MACRO> & ((callback: () => any) => FinalCommandOutput);
}
export {};
