import { isMacroArgument, } from 'sandstone/core';
import { ContainerCommandNode } from 'sandstone/core/nodes';
import { makeCallable, toMinecraftResourceName } from 'sandstone/utils';
import { coordinatesParser, rangeParser, rotationParser, targetParser, } from 'sandstone/variables/parsers';
import { CommandArguments, FinalCommandOutput } from '../../helpers.js';
import { FunctionCommandNode } from '../server/function.js';
// Yes these suck
const isObjective = (arg) => typeof arg === 'object' && Object.hasOwn(arg, 'reset');
const isScore = (arg) => typeof arg === 'object' && Object.hasOwn(arg, 'unaryOperation');
class ExecuteCommandPart extends CommandArguments {
    nestedExecute = (args, executable = true) => this.subCommand([args], (ExecuteCommand), executable);
}
export class ExecuteCommandNode extends ContainerCommandNode {
    command = 'execute';
    /**
     * Is being used by internals to generate a function call. (eg. Flow visitor)
     */
    isFake;
    /**
     * By default, the execute is treated as single (execute.run.stuff).
     * This is set to `false` if the execute is used as a function (execute.run(() => { stuff })).
     */
    isSingleExecute;
    /**
     * In case of a multiple execute, this specifies the base name of the MCFunction.
     */
    givenCallbackName;
    constructor(sandstonePack, args = [], { isFake = false, isSingleExecute = true, givenCallbackName = undefined, body = [], } = {}) {
        super(sandstonePack, ...args);
        this.givenCallbackName = givenCallbackName;
        this.isSingleExecute = isSingleExecute;
        this.isFake = isFake;
        this.append(...body);
    }
    get callbackName() {
        return this.givenCallbackName ?? `execute_${this.args[0][0].split(' ')[0]}`;
    }
    append = (...nodes) => {
        for (const node of nodes) {
            this.body.push(node);
            if (this.isSingleExecute) {
                this.sandstoneCore.getCurrentMCFunctionOrThrow().exitContext();
            }
        }
        return (nodes.length === 1 ? nodes[0] : nodes);
    };
    getValue = () => {
        if (this.body.length > 1) {
            throw new Error('Execute nodes can only have one child node when toString is called.');
        }
        // This will be the execute string without "run"
        const flattenedArgs = this.args.flat(1);
        const args = [];
        for (const arg of flattenedArgs) {
            if (arg !== undefined && arg !== null) {
                // Yes these are cursed, unfortunately, there's not really a better way to do this as visitors only visit the root nodes.
                if (typeof arg === 'object') {
                    if (isMacroArgument(this.sandstoneCore, arg)) {
                        this.isMacro = true;
                        args.push(arg.toMacro());
                    }
                    else {
                        args.push(arg);
                    }
                }
                else {
                    args.push(arg);
                }
            }
        }
        const executeString = `${this.command} ${args.join(' ')}`;
        if (this.body.length === 0) {
            return executeString;
        }
        // Yes this is cursed
        if (this.isFake) {
            return this.body[0].getValue();
        }
        let command = this.body[0].getValue();
        if (command.startsWith('$')) {
            this.isMacro = true;
            command = command.slice(1);
        }
        return `${this.isMacro ? '$' : ''}${executeString} run ${command}`;
    };
    createMCFunction = (currentMCFunction) => {
        if (this.isSingleExecute || !currentMCFunction) {
            return { node: this };
        }
        const namespace = currentMCFunction.resource.name.includes(':') ? `${currentMCFunction.resource.name.split(':')[0]}:` : '';
        // Create a new MCFunctionNode with the body of the ExecuteNode.
        const mcFunction = this.sandstonePack.MCFunction(`${namespace}${currentMCFunction.resource.path.slice(2).join('/')}/${this.callbackName}`, {
            creator: 'sandstone',
            onConflict: 'rename',
        });
        const mcFunctionNode = mcFunction['node'];
        mcFunctionNode.body = this.body;
        // Return an execute calling this MCFunction.
        const mcFunctionCall = new FunctionCommandNode(this.sandstonePack, mcFunction);
        this.body = [mcFunctionCall];
        return { node: this, mcFunction: mcFunctionNode };
    };
}
export class ExecuteStoreArgsCommand extends ExecuteCommandPart {
    /**
     * Saves the final command's return value as tag data within a block entity.
     *
     * Store as a byte, short, int, long, float, or double. If the return value is a decimal, it is rounded first and then multiplied by `scale`.
     *
     * @param targetPos Position of target block
     *
     * @param path Location of the desired tag to hold the value in.
     *
     * @param type Desired data size/type.
     *
     * @param scale Multiplier to apply before storing value. Defaults to 1.
     */
    block = (targetPos, path, type, scale) => this.nestedExecute(['block', coordinatesParser(targetPos), path, type, scale]);
    /**
     * Saves the final command's return value in either a bossbar's current value or its maximum value.
     *
     * @param id ID of the bossbar to target for saving.
     *
     * @param type Whether to overwrite the bossbar's current value or its max value.
     */
    bossbar = (id, type) => this.nestedExecute(['bossbar', id, type]);
    /**
     * Save the final command's return value in one of an entity's data tags.
     *
     * Store as a byte, short, int, long, float, or double. If the return value is a decimal, it is rounded first and then multiplied by `scale`.
     *
     * Like the `/data` command, `/execute store <arguments>` cannot modify player NBT.
     *
     * @param target A single entity to store under.
     *
     * @param path Location of the desired tag to hold the value in.
     *
     * @param type Desired data size/type.
     *
     * @param scale Multiplier to apply before storing value. Defaults to 1.
     */
    entity = (target, path, type, scale) => this.nestedExecute(['entity', targetParser(target), path, type, scale]);
    /**
     * Overrides the given score with the final command's return value.
     *
     * @param targets Specifies score holder(s) whose score is to be overridden.
     *
     * @param objective A scoreboard objective.
     *
     * @param playerScore The player's score to override.
     */
    score(...args) {
        if (isScore(args[0])) {
            return this.nestedExecute(['score', args[0]]);
        }
        return this.nestedExecute(['score', targetParser(args[0]), args[1]]);
    }
    /**
     * Uses the `path` within storage `target` to store the return value in.
     *
     * Store as a byte, short, int, long, float, or double. If the return value is a decimal, it is rounded first and then multiplied by `scale`.
     *
     * If the storage does not yet exist, it gets created.
     *
     * @param target Target storage container, as a namespaced ID.
     *
     * @param path Location of the desired tag to hold the value in.
     *
     * @param type Desired data size/type.
     *
     * @param scale Multiplier to apply before storing value. Defaults to 1.
     */
    storage = (target, path, type, scale = 1) => this.nestedExecute(['storage', target, path, type, scale]);
}
export class ExecuteStoreCommand extends ExecuteCommandPart {
    /** Store the final command's result value. */
    get result() {
        return this.subCommand([['result']], (ExecuteStoreArgsCommand), false);
    }
    /** Store the final command's success value. */
    get success() {
        return this.subCommand([['success']], (ExecuteStoreArgsCommand), false);
    }
}
export class ExecuteDataArgsCommand extends ExecuteCommandPart {
    /**
     * Checks whether the targeted block has any data for a given tag
     * @param pos Position of the block to be tested.
     * @param path Data tag to check for.
     */
    block = (pos, path) => this.nestedExecute(['block', coordinatesParser(pos), path], true);
    /**
     * Checks whether the targeted entity has any data for a given tag
     * @param target One single entity to be tested.
     * @param path Data tag to check for.
     */
    entity = (target, path) => this.nestedExecute(['entity', targetParser(target), path], true);
    /**
     * Checks whether the targeted storage has any data for a given tag
     * @param source The storage to check in.
     * @param path Data tag to check for.
     */
    storage = (source, path) => this.nestedExecute(['storage', source, path], true);
}
export class ExecuteIfUnlessCommand extends ExecuteCommandPart {
    /**
     * Compares the block at a given position to a given block.
     *
     * @param pos Position of a target block to test.
     *
     * @param block A block to test against.
     */
    block = (pos, block) => this.nestedExecute(['block', coordinatesParser(pos), block], true);
    /**
     * Checks whether one or more entities exist.
     *
     * @param targets The target entities to check.
     */
    entity = (targets) => this.nestedExecute(['entity', targetParser(targets)], true);
    /**
     * Checks whether the `predicate` evaluates to a positive result.
     *
     * @param predicate The predicate to test.
     */
    predicate(predicate) {
        if (typeof predicate === 'string') {
            return this.nestedExecute(['predicate', predicate], true);
        }
        if (Object.hasOwn(predicate, 'toMacro')) {
            return this.nestedExecute(['predicate', predicate], true);
        }
        /* @ts-ignore */
        return this.nestedExecute(['predicate', toMinecraftResourceName(predicate.path)], true);
    }
    score(...args) {
        const finalArgs = [];
        if (isScore(args[0])) {
            finalArgs.push(args[0].target.toString(), args[0].objective.name, args[1]);
            if (isScore(args[2])) {
                finalArgs.push(args[2].target.toString(), args[2].objective.name);
            }
            else {
                finalArgs.push(rangeParser(this.sandstoneCore, args[2]));
            }
        }
        else {
            finalArgs.push(targetParser(args[0]), isObjective(args[1]) ? args[1].name : args[1], args[2]);
            if (args[4]) {
                finalArgs.push(targetParser(args[3]), isObjective(args[4]) ? args[4].name : args[4]);
            }
            else {
                finalArgs.push(rangeParser(this.sandstoneCore, args[3]));
            }
        }
        return this.nestedExecute(['score', ...finalArgs], true);
    }
    /**
     * Checks whether the current dimension matches the given one.
     * @param dimension Dimension to check against.
     */
    dimension = (dimension) => this.nestedExecute(['dimension', dimension]);
    /**
     * Checks whether the targeted block is in a chunk that is fully loaded (entity-processing chunk).
     * @param pos Position of target block to test.
     */
    loaded = (pos) => this.nestedExecute(['loaded', coordinatesParser(pos)]);
    data(dataPoint) {
        if (dataPoint) {
            return this.nestedExecute(dataPoint._toMinecraftCondition().getCondition());
        }
        return this.subCommand([['data']], (ExecuteDataArgsCommand), false);
    }
    function(func) {
        if (typeof func === 'object' && Object.hasOwn(func, 'addToTag')) {
            return this.nestedExecute(['function', toMinecraftResourceName(func.path)]);
        }
        /* @ts-ignore */
        if (typeof func === 'string' || Object.hasOwn(func, 'toMacro')) {
            return this.nestedExecute(['function', func]);
        }
        const name = `${this.sandstoneCore.getCurrentMCFunctionOrThrow().resource.path.slice(2).join('/')}/execute_if_function`;
        const _func = this.sandstonePack.MCFunction(name, {
            addToSandstoneCore: false,
            creator: 'sandstone',
            onConflict: 'rename',
            callback: func,
        });
        return this.nestedExecute(['function', name]);
    }
}
export class ExecuteFacingEntityCommand extends ExecuteCommandPart {
    /**
     * Sets the command rotation to face a given point, as viewed from its anchor (either the eyes or the feet).
     *
     * @param targets The target(s) to rotate towards. (if you target multiple entities subcommands will run for each entity)
     *
     * @param anchor Whether to point at the target's eyes or feet.
     */
    entity = (targets, anchor = 'feet') => this.nestedExecute(['entity', targetParser(targets), anchor]);
}
export class ExecutePositionedCommand extends ExecuteCommandPart {
    /**
     * Sets the command position, without changing rotation or dimension, by matching an entity's position.
     *
     * @param targets Target entity/entities to match position with.
     */
    as = (targets) => this.nestedExecute(['as', targetParser(targets)]);
    /**
     * Sets the command position matching the height map (highest position in a column of blocks according to criteria) for the current position.
     *
     * - `world_surface` Any non-air block.
     * - `motion_blocking` Any motion blocking material (eg. ignores flowers and grass).
     * - `motion_blocking_no_leaves` Any non-leaf motion blocking material.
     * - `ocean_floor` Any non-fluid motion blocking material.
     */
    over = (heightMap) => this.nestedExecute(['over', heightMap]);
}
export class ExecuteRotatedAsCommand extends ExecuteCommandPart {
    /**
     * Sets the command rotation, by matching an entity's rotation.
     *
     * @param targets Target entity/entities to match rotation with.
     */
    as = (targets) => this.nestedExecute(['as', targetParser(targets)]);
}
export class ExecuteCommand extends ExecuteCommandPart {
    NodeType = ExecuteCommandNode;
    /**
     * Updates the command's position, aligning to its current block position (an integer). Only applies along specified axes.
     * This is akin to flooring the coordinates – i.e. rounding them downwards. It updates the meaning of `~ ~ ~` and `^ ^ ^`.
     *
     * @param axes Any non-repeating combination of the characters 'x', 'y', and 'z'.
     */
    align = (axes) => this.nestedExecute(['align', axes]);
    /**
     * Stores the distance from the feet to the eyes of the entity that is executing the command in the anchor, which is part of the command context.
     * Effectively re-centers `^ ^ ^` on either the eyes or feet, also changing the angle the `facing` sub-command works off of.
     *
     * @param anchor Whether to anchor the executed command to eye level or feet level
     */
    anchored = (anchor) => this.nestedExecute(['anchored', anchor]);
    /**
     * Sets the command sender to target entity, without changing position, rotation, dimension, or anchor
     *
     * @param targets Target entity/entities to become the new sender.
     */
    as = (targets) => this.nestedExecute(['as', targetParser(targets)]);
    /**
     * Sets the command position, rotation, and dimension to match those of an entity/entities; does not change sender
     * @param targets Target entity/entities to match position, rotation, and dimension with
     */
    at = (targets) => this.nestedExecute(['at', targetParser(targets)]);
    facing(pos) {
        if (pos) {
            return this.nestedExecute(['facing', coordinatesParser(pos)]);
        }
        return this.subCommand([['facing']], (ExecuteFacingEntityCommand), false);
    }
    /**
     * Sets the command dimension.
     *
     * @param dimension Name of the new execution dimension.
     */
    in = (dimension) => this.nestedExecute(['in', dimension]);
    /**
     * Sets the command executor to a related entity.
     * @param relation Relation with the current executor.
     */
    on = (relation) => this.nestedExecute(['on', relation]);
    positioned(pos) {
        if (pos) {
            return this.nestedExecute(['positioned', coordinatesParser(pos)]);
        }
        return this.subCommand([['positioned']], (ExecutePositionedCommand), false);
    }
    rotated(rotation) {
        if (rotation) {
            return this.nestedExecute(['rotated', rotationParser(rotation)]);
        }
        return this.subCommand([['rotated']], ExecuteRotatedAsCommand, false);
    }
    /**
     * Summons an entity at the command position and sets the command executor to said newly summoned entity.
     * @param entityType Entity to summon.
     */
    summon(entityType) {
        return this.nestedExecute(['summon', entityType]);
    }
    get if() {
        return this.subCommand([['if']], (ExecuteIfUnlessCommand), false);
    }
    get unless() {
        return this.subCommand([['unless']], (ExecuteIfUnlessCommand), false);
    }
    get store() {
        return this.subCommand([['store']], (ExecuteStoreCommand), false);
    }
    get run() {
        const node = this.getNode();
        const commands = new Proxy(this.sandstonePack.commands, {
            get: (target, p, receiver) => {
                // The context will automatically be exited by the node itself
                this.sandstoneCore.getCurrentMCFunctionOrThrow().enterContext(node, false);
                return this.sandstonePack.commands[p];
            },
        });
        return makeCallable(commands, (callback) => {
            node.isSingleExecute = false;
            this.sandstoneCore.insideContext(node, callback, false);
            return new FinalCommandOutput(node);
        }, true);
    }
}
