import { CommandNode } from 'sandstone/core/nodes';
import { CommandArguments } from '../../helpers.js';
import type { DISPLAY_SLOTS, FormattingTags, JSONTextComponent, MultipleEntitiesArgument, OBJECTIVE_CRITERION, ObjectiveArgument, OPERATORS } from 'sandstone/arguments';
import type { Macroable } from 'sandstone/core';
import type { LiteralUnion } from 'sandstone/utils';
import type { Score } from 'sandstone/variables';
export declare class ScoreboardCommandNode extends CommandNode {
    command: "scoreboard";
}
export declare class ScoreboardObjectivesModifyCommand<MACRO extends boolean> extends CommandArguments {
    /**
     * Change the display name of the scoreboard in display slots.
     *
     * @param displayName The new display name. Must be a JSON text component.
     */
    displayname: (displayName?: Macroable<JSONTextComponent, MACRO>) => import("../../helpers.js").FinalCommandOutput;
    /**
     * Change the display format of health bars.
     *
     * @param display Whether to display the health bars as hearts or integers.
     */
    rendertype: (display: Macroable<'hearts' | 'integer', MACRO>) => import("../../helpers.js").FinalCommandOutput;
}
export declare class ScoreboardCommand<MACRO extends boolean> extends CommandArguments {
    protected NodeType: typeof ScoreboardCommandNode;
    /** All commands related to scoreboard objectives. */
    objectives: {
        /** List all existing objectives with their display names and criteria. */
        list: () => import("../../helpers.js").FinalCommandOutput;
        /**
         * Create a new objective with the given internal objective name, specified criterion, and the optional display name.
         * All three arguments are case-sensitive.
         *
         * @param objective must be a plain text.
         *
         * @param criterion must be a valid criterion type.
         *
         * @param displayName must be a JSON text component, defaulting to `objective` when unspecified.
         */
        add: (objective: Macroable<ObjectiveArgument, MACRO>, criteria: Macroable<LiteralUnion<OBJECTIVE_CRITERION>, MACRO>, displayName?: Macroable<JSONTextComponent, MACRO>) => import("../../helpers.js").FinalCommandOutput;
        /**
         * Delete all references to the named objective in the scoreboard system.
         * Data is deleted from the objectives list and entity scores,
         * and if it was on a display list it is no longer displayed
         */
        remove: (objective: Macroable<ObjectiveArgument, MACRO>) => import("../../helpers.js").FinalCommandOutput;
        /**
         * Display score info for the objective in the given slot.
         *
         * @param slot The slot to display the objective in.
         *
         * @param objective The objective to display. If not provided, this display slot is cleared.
         */
        setdisplay: (slot: Macroable<DISPLAY_SLOTS, MACRO>, objective?: Macroable<ObjectiveArgument, MACRO>) => import("../../helpers.js").FinalCommandOutput;
        /**
         * Modify the display of the objective.
         *
         * @param objective The objective to modify.
         */
        modify: (objective: Macroable<ObjectiveArgument, MACRO>) => ScoreboardObjectivesModifyCommand<boolean>;
    };
    /** All commands related to scoreboard players. */
    players: {
        /**
         * If `target` is omitted, lists all entities which are tracked in some way by the scoreboard system.
         *
         * Else, lists the scores of a particular entity.
         *
         * @param target The entity to list the scores from.
         */
        readonly list: (target?: Macroable<MultipleEntitiesArgument<MACRO>, MACRO>) => import("../../helpers.js").FinalCommandOutput;
        /**
         * Return the scoreboard value of a given objective for a given target.
         *
         * @param target The entity to get the score from.
         *
         * @param objective The objective to get the score from.
         */
        readonly get: (...args: [target: Macroable<MultipleEntitiesArgument<MACRO> | number, MACRO>, objective: Macroable<ObjectiveArgument, MACRO>] | [target: Macroable<Score, MACRO>]) => import("../../helpers.js").FinalCommandOutput;
        readonly set: (...args: [..._: ([target: Macroable<MultipleEntitiesArgument<MACRO> | number, MACRO>, objective: Macroable<ObjectiveArgument, MACRO>] | [target: Macroable<Score, MACRO>]), score: Macroable<number, MACRO>]) => import("../../helpers.js").FinalCommandOutput;
        readonly add: (...args: [..._: ([target: Macroable<MultipleEntitiesArgument<MACRO> | number, MACRO>, objective: Macroable<ObjectiveArgument, MACRO>] | [target: Macroable<Score, MACRO>]), score: Macroable<number, MACRO>]) => import("../../helpers.js").FinalCommandOutput;
        readonly remove: (...args: [..._: ([target: Macroable<MultipleEntitiesArgument<MACRO> | number, MACRO>, objective: Macroable<ObjectiveArgument, MACRO>] | [target: Macroable<Score, MACRO>]), score: Macroable<number, MACRO>]) => import("../../helpers.js").FinalCommandOutput;
        readonly reset: (...args: [target: Macroable<MultipleEntitiesArgument<MACRO> | number, MACRO>, objective: Macroable<ObjectiveArgument, MACRO>] | [target: Macroable<Score, MACRO>]) => import("../../helpers.js").FinalCommandOutput;
        readonly enable: (...args: [target: Macroable<MultipleEntitiesArgument<MACRO> | number, MACRO>, objective: Macroable<ObjectiveArgument, MACRO>] | [target: Macroable<Score, MACRO>]) => import("../../helpers.js").FinalCommandOutput;
        readonly operation: (...args: [
            ...target: [target: Macroable<MultipleEntitiesArgument<MACRO> | number, MACRO>, targetObjective: Macroable<ObjectiveArgument, MACRO>] | [targetScore: Macroable<Score, MACRO>],
            operation: OPERATORS,
            ...source: [source: Macroable<MultipleEntitiesArgument<MACRO> | number, MACRO>, sourceObjective: Macroable<ObjectiveArgument, MACRO>] | [sourceScore: Macroable<Score, MACRO>]
        ]) => import("../../helpers.js").FinalCommandOutput;
        readonly display: {
            readonly name: (...args: [
                ...target: [target: Macroable<MultipleEntitiesArgument<MACRO> | number, MACRO>, targetObjective: Macroable<ObjectiveArgument, MACRO>] | [targetScore: Macroable<Score, MACRO>],
                name: Macroable<JSONTextComponent, MACRO>
            ]) => import("../../helpers.js").FinalCommandOutput;
            readonly numberformat: (...args: [
                ...target: [target: Macroable<MultipleEntitiesArgument<MACRO> | number, MACRO>, targetObjective: Macroable<ObjectiveArgument, MACRO>] | [targetScore: Macroable<Score, MACRO>],
                ...format: [format: 'blank'] | [format: 'styled', style: Macroable<FormattingTags, MACRO>] | [format: 'fixed', rightColumn: Macroable<JSONTextComponent, MACRO>]
            ]) => import("../../helpers.js").FinalCommandOutput;
        };
    };
}
