import { CommandNode } from 'sandstone/core/nodes';
import { CommandArguments } from '../../helpers.js';
import type { Coordinates, NBTObject, SingleEntityArgument } from 'sandstone/arguments';
import type { Macroable } from 'sandstone/core';
export declare class DataCommandNode extends CommandNode {
    command: "data";
}
/** Allows to get, merge, modify, and remove NBT data of a block entity, entity, or Command NBT storage. */
export declare class DataCommand<MACRO extends boolean> extends CommandArguments {
    protected NodeType: typeof DataCommandNode;
    /** Read off the entire NBT data or the subsection of the NBT data from the targeted block position or entity, scaled by `scale` if specified. */
    get get(): DataGetCommand<MACRO>;
    /** Merge the NBT data from the sourced block position or entity with the specified `nbt` data. */
    get merge(): DataMergeCommand<MACRO>;
    /** Modify the NBT data from the sourced block position or entity, with the specified operation and the given NBT. */
    get modify(): DataModifyCommand<MACRO>;
    /** Removes NBT data at `path` from the targeted block position or entity. Player NBT data cannot be removed. */
    get remove(): DataRemoveCommand<MACRO>;
}
export declare class DataGetCommand<MACRO extends boolean> extends CommandArguments {
    /**
     * Get the NBT of a block at the given position.
     *
     * @param targetPos The coordinates of the block to get the NBT from.
     * @param path The path of the NBT to get.
     * @param scale The scale to multiply the NBT value by.
     */
    block: (targetPos: Macroable<Coordinates<MACRO>, MACRO>, path?: Macroable<string, MACRO>, scale?: Macroable<number, MACRO>) => import("../../helpers.js").FinalCommandOutput;
    /**
     * Get the NBT of a given entity.
     *
     * @param target The entity to get the NBT from.
     * @param path The path of the NBT to get.
     * @param scale The scale to multiply the NBT value by.
     */
    entity: (target: Macroable<SingleEntityArgument<MACRO>, MACRO>, path?: Macroable<string, MACRO>, scale?: Macroable<number, MACRO>) => import("../../helpers.js").FinalCommandOutput;
    /**
     * Get the NBT from a given storage path.
     *
     * @param target The storage to get the NBT from.
     * @param path The path of the NBT to get.
     * @param scale The scale to multiply the NBT value by.
     */
    storage: (target: Macroable<string, MACRO>, path?: Macroable<string, MACRO>, scale?: Macroable<number, MACRO>) => import("../../helpers.js").FinalCommandOutput;
}
export declare class DataMergeCommand<MACRO extends boolean> extends CommandArguments {
    /**
     * Merge the NBT of a block at the given position, with the given NBT.
     *
     * @param targetPos The coordinates of the block to merge the NBT with.
     * @param nbt The NBT to merge with.
     */
    block: (targetPos: Macroable<Coordinates<MACRO>, MACRO>, nbt: Macroable<NBTObject, MACRO>) => import("../../helpers.js").FinalCommandOutput;
    /**
     * Merge the NBT of the given entity, with the given NBT.
     *
     * @param target The entity to merge the NBT with.
     * @param nbt The NBT to merge with.
     */
    entity: (target: Macroable<SingleEntityArgument<MACRO>, MACRO>, nbt: Macroable<NBTObject, MACRO>) => import("../../helpers.js").FinalCommandOutput;
    /**
     * Merge the NBT of the given storage path, with the given NBT.
     *
     * @param target The storage to merge the NBT with.
     * @param nbt The NBT to merge with.
     */
    storage: (target: Macroable<string, MACRO>, nbt: Macroable<NBTObject, MACRO>) => import("../../helpers.js").FinalCommandOutput;
}
export declare class DataModifyValuesCommand<MACRO extends boolean> extends CommandArguments {
    from: {
        /**
         * Modify with the NBT of a block at the given position.
         *
         * @param sourcePosition The coordinates of the block to modify the NBT with.
         * @param sourcePath The path of the NBT to modify with.
         */
        block: (sourcePosition: Macroable<Coordinates<MACRO>, MACRO>, sourcePath: Macroable<string, MACRO>) => import("../../helpers.js").FinalCommandOutput;
        /**
         * Modify with the NBT of a given entity.
         *
         * @param source The entity to modify the NBT with.
         * @param sourcePath The path of the NBT to modify with.
         */
        entity: (source: Macroable<SingleEntityArgument<MACRO>, MACRO>, sourcePath: Macroable<string, MACRO>) => import("../../helpers.js").FinalCommandOutput;
        /**
         * Modify with the NBT of a given storage path.
         *
         * @param source The storage path to modify the NBT with.
         * @param sourcePath The path of the NBT to modify with.
         */
        storage: (source: Macroable<string, MACRO>, sourcePath: Macroable<string, MACRO>) => import("../../helpers.js").FinalCommandOutput;
    };
    string: {
        /**
         * Modify with the NBT String of a block at the given position.
         *
         * @param sourcePosition The coordinates of the block to modify the NBT with.
         * @param sourcePath The path of the NBT to modify with.
         * @param start Optional. Index of first character to include at the start of the string.
         * @param end Optional. Index of the first character to exclude at the end of the string
         */
        block: (sourcePosition: Macroable<Coordinates<MACRO>, MACRO>, sourcePath: Macroable<string, MACRO>, start?: Macroable<number, MACRO>, end?: Macroable<number, MACRO>) => void;
        /**
         * Modify with the NBT String of a given entity.
         *
         * @param source The entity to modify the NBT with.
         * @param sourcePath The path of the NBT to modify with.
         * @param start Optional. Index of first character to include at the start of the string.
         * @param end Optional. Index of the first character to exclude at the end of the string
         */
        entity: (source: Macroable<SingleEntityArgument<MACRO>, MACRO>, sourcePath: Macroable<string, MACRO>, start?: Macroable<number, MACRO>, end?: Macroable<number, MACRO>) => void;
        /**
         * Modify with the NBT String of a given storage path.
         *
         * @param source The storage path to modify the NBT with.
         * @param sourcePath The path of the NBT to modify with.
         * @param start Optional. Index of first character to include at the start of the string.
         * @param end Optional. Index of the first character to exclude at the end of the string
         */
        storage: (source: Macroable<string, MACRO>, sourcePath: Macroable<string, MACRO>, start?: Macroable<number, MACRO>, end?: Macroable<number, MACRO>) => void;
    };
    /**
     * Modify the NBT with the given value.
     */
    value: (value: Macroable<NBTObject, MACRO>) => import("../../helpers.js").FinalCommandOutput;
}
export declare class DataModifyTypeCommand<MACRO extends boolean> extends CommandArguments {
    /** Append the source data onto the end of the pointed-to list. */
    get append(): DataModifyValuesCommand<MACRO>;
    /**
     * Insert the source data into the pointed-to list as element `index`, then shift higher elements one position upwards.
     *
     * @param index The index to insert the NBT to.
     */
    insert: (index: Macroable<number, MACRO>) => DataModifyValuesCommand<MACRO>;
    /** Merge the source data into the pointed-to object. */
    get merge(): DataModifyValuesCommand<MACRO>;
    /** Prepend the source data onto the beginning of the pointed-to list. */
    get prepend(): DataModifyValuesCommand<MACRO>;
    /** Set the tag specified by `targetPath` to the source data. */
    get set(): DataModifyValuesCommand<MACRO>;
}
export declare class DataModifyCommand<MACRO extends boolean> extends CommandArguments {
    /**
     * Modify the NBT of a block at the given position.
     *
     * @param targetPos The coordinates of the block to modify the NBT from.
     * @param path The path of the NBT to modify.
     */
    block: (targetPos: Macroable<Coordinates<MACRO>, MACRO>, targetPath: Macroable<string, MACRO>) => DataModifyTypeCommand<MACRO>;
    /**
     * Modify the NBT of a given entity.
     *
     * @param target The entity to modify the NBT from.
     * @param path The path of the NBT to modify.
     */
    entity: (target: Macroable<SingleEntityArgument<MACRO>, MACRO>, targetPath: Macroable<string, MACRO>) => DataModifyTypeCommand<MACRO>;
    /**
     * Modify the NBT from a given storage path.
     *
     * @param target The storage to modify the NBT from.
     * @param path The path of the NBT to modify.
     */
    storage: (target: Macroable<string, MACRO>, targetPath: Macroable<string, MACRO>) => DataModifyTypeCommand<MACRO>;
}
export declare class DataRemoveCommand<MACRO extends boolean> extends CommandArguments {
    /**
     * Remove the NBT of a block at the given position.
     *
     * @param targetPos The coordinates of the block to remove the NBT from.
     * @param path The path of the NBT to remove.
     */
    block: (targetPos: Macroable<Coordinates<MACRO>, MACRO>, targetPath: Macroable<string, MACRO>) => import("../../helpers.js").FinalCommandOutput;
    /**
     * Remove the NBT of a given entity.
     *
     * @param target The entity to remove the NBT from.
     * @param path The path of the NBT to remove.
     */
    entity: (target: Macroable<SingleEntityArgument<MACRO>, MACRO>, targetPath: Macroable<string, MACRO>) => import("../../helpers.js").FinalCommandOutput;
    /**
     * Remove the NBT from a given storage path.
     *
     * @param target The storage to remove the NBT from.
     * @param path The path of the NBT to remove.
     */
    storage: (target: Macroable<string, MACRO>, targetPath: Macroable<string, MACRO>) => import("../../helpers.js").FinalCommandOutput;
}
