/* eslint-disable @typescript-eslint/no-non-null-assertion */
import AdmZip from 'adm-zip';
import fs from 'fs-extra';
import lodash from 'lodash';
import path from 'path';
import { fetch, safeWrite } from 'sandstone/utils';
export class MCMetaCache {
    base = '/misode/mcmeta/';
    path = path.join(process.env.WORKING_DIR, '..', 'resources', 'cache', 'mcmeta');
    manifest = path.join(this.path, '..', '..', 'mcmeta.json');
    lockFile = path.join(this.path, '..', 'lock-mcmeta.json');
    version = 'latest';
    versionDate;
    branches = [];
    files = new Map();
    loaded = false;
    async load() {
        this.loaded = true;
        if (!(await fs.pathExists(this.manifest))) {
            return;
        }
        const manifest = JSON.parse(await fs.readFile(this.manifest, 'utf-8'));
        let uncached = false;
        this.version = manifest.version;
        const getZip = async (branch) => new AdmZip(await (await fetch(`https://github.com${this.base}archive/refs/heads/${branch}.zip`)).buffer());
        if (!(await fs.pathExists(this.lockFile))) {
            if (manifest.files.length > 10) {
                const archives = {};
                for await (const branch of manifest.branches) {
                    archives[branch] = await getZip(branch);
                }
                for await (const [file, text] of Object.entries(manifest.files)) {
                    const _path = file.split('/');
                    let contents = await new Promise((res) => {
                        archives[_path[0]].readFileAsync(_path.slice(1).join('/'), (data) => res(data));
                    });
                    if (text) {
                        contents = contents.toString('utf-8');
                    }
                    await safeWrite(path.join(this.path, file), contents);
                    this.files.set(file, { contents, text: text });
                }
            }
            else {
                uncached = true;
                for await (const [file, text] of Object.entries(manifest.files)) {
                    const _path = file.split('/');
                    await this.get(_path[0], _path.slice(1).join('/'), text);
                }
            }
            await safeWrite(this.lockFile, JSON.stringify({
                files: manifest.files,
                version: manifest.version === 'latest' ? await this.getVersion(false) : manifest.version,
                versionDate: this.versionDate,
            }));
        }
        else {
            const lockFile = JSON.parse(await fs.readFile(this.lockFile, 'utf-8'));
            this.versionDate = lockFile.versionDate;
            const version = manifest.version === 'latest' ? await this.getVersion(lockFile) : manifest.version;
            if (lockFile.version !== version) {
                if (manifest.files.length > 10) {
                    const archives = {};
                    for await (const branch of manifest.branches) {
                        archives[branch] = await getZip(branch);
                    }
                    for await (const [file, text] of Object.entries(manifest.files)) {
                        const _path = file.split('/');
                        let contents = await new Promise((res) => {
                            archives[_path[0]].readFileAsync(_path.slice(1).join('/'), (data) => res(data));
                        });
                        if (text) {
                            contents = contents.toString('utf-8');
                        }
                        await safeWrite(path.join(this.path, file), contents);
                        this.files.set(file, { contents, text: text });
                    }
                }
                else {
                    uncached = true;
                    for await (const [file, text] of Object.entries(manifest.files)) {
                        const _path = file.split('/');
                        await this.get(_path[0], _path.slice(1).join('/'), text);
                    }
                }
            }
            else {
                const diff = lodash.difference(Object.entries(manifest.files), Object.entries(lockFile.files));
                if (diff.length !== 0) {
                    if (diff.length > 10) {
                        const archives = {};
                        for await (const [file, text] of Object.entries(manifest.files)) {
                            const _path = file.split('/');
                            const branch = _path[0];
                            if (!archives[branch]) {
                                archives[branch] = await getZip(branch);
                            }
                            let contents = await new Promise((res) => {
                                archives[branch].readFileAsync(_path.slice(1).join('/'), (data) => res(data));
                            });
                            if (text) {
                                contents = contents.toString('utf-8');
                            }
                            await safeWrite(path.join(this.path, file), contents);
                            this.files.set(file, { contents, text: text });
                        }
                    }
                    else {
                        uncached = true;
                        for await (const [file, text] of diff) {
                            const _path = file.split('/');
                            await this.get(_path[0], _path.slice(1).join('/'), text);
                        }
                    }
                }
            }
            await safeWrite(this.lockFile, JSON.stringify({
                files: manifest.files,
                version: manifest.version === 'latest' ? await this.getVersion(false) : manifest.version,
                versionDate: this.versionDate,
            }));
        }
        this.branches = manifest.branches;
        if (!uncached) {
            for (const [file, text] of Object.entries(manifest.files)) {
                this.files.set(file, { contents: false, text: text });
            }
        }
    }
    async save() {
        if (this.files.size !== 0) {
            const files = {};
            for (const [_path, { text }] of this.files.entries()) {
                files[_path] = text;
            }
            await safeWrite(this.manifest, JSON.stringify({
                files,
                branches: this.branches,
                version: this.version,
            }));
            await safeWrite(this.lockFile, JSON.stringify({
                files,
                version: this.version,
                versionDate: this.versionDate || Date.now(),
            }));
        }
    }
    async get(branch, relativePath, text = true) {
        console.log(branch, relativePath);
        if (!this.loaded) {
            await this.load();
        }
        if (!this.branches.includes(branch)) {
            this.branches.push(branch);
        }
        const fullPath = path.join(branch, relativePath);
        const existing = this.files.get(fullPath);
        if (existing) {
            return existing.contents;
        }
        const req = await fetch(`https://raw.githubusercontent.com${this.base}${branch}/${relativePath}`);
        const file = await (text ? req.text() : req.buffer());
        const finalPath = path.join(this.path, fullPath);
        await safeWrite(finalPath, file);
        return (this.files.set(fullPath, { text, contents: file }).get(fullPath)).contents;
    }
    async getVersion(lockFile) {
        const fetchVersion = async () => (await (await fetch(`https://api.github.com/repos${this.base}commits?sha=summary`)).json()).sha;
        const currentDate = Date.now();
        if (lockFile) {
            // If it has been over 6 hours since it was checked.
            if (((lockFile.versionDate - currentDate) / 36e5) > 6) {
                this.versionDate = currentDate;
                return fetchVersion();
            }
            this.versionDate = lockFile.versionDate;
            return lockFile.version;
        }
        this.versionDate = currentDate;
        return fetchVersion();
    }
}
