import { ContainerNode } from '../../nodes.js';
import { ResourceClass } from '../resource.js';
import type { AdvancementJSON, MultiplePlayersArgument } from 'sandstone/arguments';
import type { ConditionClass } from 'sandstone/variables';
import type { SandstoneCore } from '../../sandstoneCore.js';
import type { ResourceClassArguments, ResourceNode } from '../resource.js';
/**
 * A node representing a Minecraft advancement.
 */
export declare class AdvancementNode extends ContainerNode implements ResourceNode<AdvancementClass> {
    resource: AdvancementClass;
    constructor(sandstoneCore: SandstoneCore, resource: AdvancementClass);
    getValue: () => string;
}
export type AdvancementClassArguments<CriteriaNames extends string = string> = {
    /**
     * The advancement's JSON.
     */
    advancement?: AdvancementJSON<CriteriaNames>;
} & ResourceClassArguments<'default'>;
export declare class AdvancementClass<CriteriaNames extends string = string> extends ResourceClass<AdvancementNode> implements ConditionClass {
    advancementJSON: NonNullable<AdvancementClassArguments['advancement']>;
    constructor(sandstoneCore: SandstoneCore, name: string, args: AdvancementClassArguments<CriteriaNames>);
    /**
     * Grant this advancement.
     * @param players Optional. Specifies the player(s). Defaults to `@s`.
     */
    grant(players?: MultiplePlayersArgument<false>, criterion?: CriteriaNames): void;
    /**
     * Revoke this advancement.
     * @param players Optional. Specifies the player(s). Defaults to `@s`.
     */
    revoke(players?: MultiplePlayersArgument<false>, criterion?: CriteriaNames): void;
    /**
     * Grant this advancement and all its parent advancements.
     * Think of specifying everything from the start *until* that advancement.
     *
     * The exact order the operation is carried out in is: `parent > parent's parent > ... > root > this advancement.`
     *
     * @param players Optional. Specifies the player(s). Defaults to `@s`.
     */
    grantUntilThis(players?: MultiplePlayersArgument<false>): void;
    /**
     * Revoke this advancement and all its parent advancements.
     * Think of specifying everything from the start *until* that advancement.
     *
     * The exact order the operation is carried out in is: `parent > parent's parent > ... > root > this advancement.`
     *
     * @param players Optional. Specifies the player(s). Defaults to `@s`.
     */
    revokeUntilThis(players?: MultiplePlayersArgument<false>): void;
    /**
     * Grant this advancement and all its children advancements.
     * Think of specifying everything *from* that advancement to the end.
     *
     * The exact order the operation is carried out in is specified `advancement > child > child's child > ...`.
     * When it operates on a child that branches, it iterates through all its children before continuing.
     *
     * @param players Optional. Specifies the player(s). Defaults to `@s`.
     */
    grantFromThis(players?: MultiplePlayersArgument<false>): void;
    /**
     * Revoke this advancement and all its children advancements.
     * Think of specifying everything *from* that advancement to the end.
     *
     * The exact order the operation is carried out in is specified `advancement > child > child's child > ...`.
     * When it operates on a child that branches, it iterates through all its children before continuing.
     *
     * @param players Optional. Specifies the player(s). Defaults to `@s`.
     */
    revokeFromThis(players?: MultiplePlayersArgument<false>): void;
    /**
     * Grant this advancement, all its parent advancements, and all its children advancements.
     * Think of specifying everything through the specified advancement, going both backwards and forwards.
     *
     * The exact order the operation is as if the command were executed with `until` specified, then with `from` specified:
     * `parent > parent's parent > ... > root > specified advancement > child > child's child > ...`
     *
     * @param players Optional. Specifies the player(s). Defaults to `@s`.
     */
    grantThroughThis(players?: MultiplePlayersArgument<false>): void;
    /**
     * Revoke this advancement, all its parent advancements, and all its children advancements.
     * Think of specifying everything through the specified advancement, going both backwards and forwards.
     *
     * The exact order the operation is as if the command were executed with `until` specified, then with `from` specified:
     * `parent > parent's parent > ... > root > specified advancement > child > child's child > ...`
     *
     * @param players Optional. Specifies the player(s). Defaults to `@s`.
     */
    revokeThroughThis(players?: MultiplePlayersArgument<false>): void;
    /**
     * @internal
     */
    _toMinecraftCondition: () => import("../../../flow/conditions/resources/advancement.js").AdvancementConditionNode;
    /**
     * Test if the currently executing player has a specific advancement criterion.
     */
    criterion(criterion: string): import("../../../flow/conditions/resources/advancement.js").AdvancementConditionNode;
}
