import { ContainerNode } from '../../nodes.js';
import { ResourceClass } from '../resource.js';
/**
 * A node representing a Minecraft advancement.
 */
export class AdvancementNode extends ContainerNode {
    resource;
    constructor(sandstoneCore, resource) {
        super(sandstoneCore);
        this.resource = resource;
    }
    getValue = () => JSON.stringify(this.resource.advancementJSON);
}
export class AdvancementClass extends ResourceClass {
    advancementJSON;
    constructor(sandstoneCore, name, args) {
        super(sandstoneCore, { packType: sandstoneCore.pack.dataPack(), extension: 'json' }, AdvancementNode, sandstoneCore.pack.resourceToPath(name, ['advancements']), args);
        this.advancementJSON = args.advancement;
        this.handleConflicts();
    }
    /**
     * Grant this advancement.
     * @param players Optional. Specifies the player(s). Defaults to `@s`.
     */
    grant(players = '@s', criterion) {
        this.pack.commands.advancement.grant(players).only(this.name, criterion);
    }
    /**
     * Revoke this advancement.
     * @param players Optional. Specifies the player(s). Defaults to `@s`.
     */
    revoke(players = '@s', criterion) {
        this.pack.commands.advancement.revoke(players).only(this.name, criterion);
    }
    /**
     * Grant this advancement and all its parent advancements.
     * Think of specifying everything from the start *until* that advancement.
     *
     * The exact order the operation is carried out in is: `parent > parent's parent > ... > root > this advancement.`
     *
     * @param players Optional. Specifies the player(s). Defaults to `@s`.
     */
    grantUntilThis(players = '@s') {
        this.pack.commands.advancement.grant(players).until(this.name);
    }
    /**
     * Revoke this advancement and all its parent advancements.
     * Think of specifying everything from the start *until* that advancement.
     *
     * The exact order the operation is carried out in is: `parent > parent's parent > ... > root > this advancement.`
     *
     * @param players Optional. Specifies the player(s). Defaults to `@s`.
     */
    revokeUntilThis(players = '@s') {
        this.pack.commands.advancement.revoke(players).until(this.name);
    }
    /**
     * Grant this advancement and all its children advancements.
     * Think of specifying everything *from* that advancement to the end.
     *
     * The exact order the operation is carried out in is specified `advancement > child > child's child > ...`.
     * When it operates on a child that branches, it iterates through all its children before continuing.
     *
     * @param players Optional. Specifies the player(s). Defaults to `@s`.
     */
    grantFromThis(players = '@s') {
        this.pack.commands.advancement.grant(players).from(this.name);
    }
    /**
     * Revoke this advancement and all its children advancements.
     * Think of specifying everything *from* that advancement to the end.
     *
     * The exact order the operation is carried out in is specified `advancement > child > child's child > ...`.
     * When it operates on a child that branches, it iterates through all its children before continuing.
     *
     * @param players Optional. Specifies the player(s). Defaults to `@s`.
     */
    revokeFromThis(players = '@s') {
        this.pack.commands.advancement.revoke(players).from(this.name);
    }
    /**
     * Grant this advancement, all its parent advancements, and all its children advancements.
     * Think of specifying everything through the specified advancement, going both backwards and forwards.
     *
     * The exact order the operation is as if the command were executed with `until` specified, then with `from` specified:
     * `parent > parent's parent > ... > root > specified advancement > child > child's child > ...`
     *
     * @param players Optional. Specifies the player(s). Defaults to `@s`.
     */
    grantThroughThis(players = '@s') {
        this.pack.commands.advancement.grant(players).through(this.name);
    }
    /**
     * Revoke this advancement, all its parent advancements, and all its children advancements.
     * Think of specifying everything through the specified advancement, going both backwards and forwards.
     *
     * The exact order the operation is as if the command were executed with `until` specified, then with `from` specified:
     * `parent > parent's parent > ... > root > specified advancement > child > child's child > ...`
     *
     * @param players Optional. Specifies the player(s). Defaults to `@s`.
     */
    revokeThroughThis(players = '@s') {
        this.pack.commands.advancement.revoke(players).through(this.name);
    }
    /**
     * @internal
     */
    _toMinecraftCondition = () => new this.pack.conditions.Advancement(this.core, this.name);
    /**
     * Test if the currently executing player has a specific advancement criterion.
     */
    criterion(criterion) {
        return new this.pack.conditions.Advancement(this.core, this.name, criterion);
    }
}
