import { ContainerNode } from '../../nodes.js';
import { CallableResourceClass } from '../resource.js';
import { TagClass } from './tag.js';
import type { TimeArgument } from 'sandstone/arguments/basics';
import type { ScheduleType } from 'sandstone/commands';
import type { FinalCommandOutput } from 'sandstone/commands/helpers';
import type { ContainerCommandNode, MacroArgument, Node, ResourceClassArguments, ResourceNode, SandstoneCore } from 'sandstone/core';
import type { MakeInstanceCallable } from 'sandstone/utils';
/**
 * A node representing a Minecraft function.
 */
export declare class MCFunctionNode extends ContainerNode implements ResourceNode {
    resource: MCFunctionClass<any, any>;
    contextStack: (ContainerNode | ContainerCommandNode)[];
    constructor(sandstoneCore: SandstoneCore, resource: MCFunctionClass<any, any>);
    /**
     * The currently active context.
     *
     * For example, the current context is the function body if the function is not in a loop.
     * If the function is in a loop, the current context is the loop body.
     */
    get currentContext(): ContainerNode | ContainerCommandNode<unknown[]>;
    /**
     * Sequentially add node(s) to the end of the body of the function.
     *
     * @param node The node(s) to add.
     */
    appendNode: (node: Node | Node[]) => void;
    /**
     * Sequentially add node(s) to the beginning of the body of the function.
     *
     * @param node The node(s) to add.
     */
    prependNode: (node: Node | Node[]) => void;
    /**
     * Switch the current context to the given node.
     * Also adds the node to the body of the current context, except if addNode is False.
     *
     * @param node The node to switch to.
     * @param addNode Whether to add the node to the body of the current context.
     */
    enterContext: (node: ContainerNode | ContainerCommandNode, addNode?: boolean) => void;
    /**
     * Switch the current context to the given node, run the given function, and switch back to the original context.
     * Also adds the node to the body of the current context, except if addNode is False.
     *
     * @param node The node to switch to.
     * @param callback The function to run.
     * @param addNode Whether to add the node to the body of the current context.
     *
     * @return The previously active context.
     * @throws Error if there is no previous context.
     */
    insideContext: (node: ContainerNode | ContainerCommandNode, callback: () => void, addNode?: boolean) => ContainerNode | ContainerCommandNode<unknown[]> | undefined;
    /**
     * Leave the current context, and return to the previous one.
     *
     * @return The previously active context.
     * @throws Error if there is no previous context.
     */
    exitContext: () => ContainerNode | ContainerCommandNode<unknown[]> | undefined;
    getValue: () => string;
}
export type MCFunctionClassArguments = ({
    /**
     * The callback to run when the MCFunction is generated.
     *
     * @default () => {}
     */
    callback?: (...params: [loop: MCFunctionClass<any, any>, ...params: MacroArgument[]]) => void;
    /**
     * If true, then the function will only be created if it is called from another function. TODO: implement this
     */
    lazy?: boolean;
    /**
     * Whether the function should run when the datapack loads.
     *
     * Defaults to `true` if `runEvery` is specified, else `false`.
     *
     * Incompatible with macro parameters.
     */
    runOnLoad?: boolean;
    /**
     * The function tags to put this function in.
     */
    tags?: readonly (string | TagClass<'functions'>)[];
    /**
     * If specified, the function will run every given time.
     *
     * If `runOnLoad` is unspecified or `true`, then it will run on load too.
     *
     * If `runOnLoad` is `false`, you will have to manually start it.
     *
     * You can stop the automatic scheduling by running `theFunction.clearSchedule()`.
     *
     * Incompatible with macro parameters.
     *
     * @example
     *
     * // Run every 5 ticks, including on datapack load.
     * {
     *   runEvery: 5,
     * }
     *
     * // Run every 5 ticks, but wait until 5 ticks after the datapack loads for 1st execution.
     * {
     *   runEvery: 5,
     *   runOnLoad: false,
     * }
     *
     * // Run every 8 seconds
     * {
     *   runEvery: '8s'
     * }
     */
    runEvery?: TimeArgument;
    /**
     * Whether the function should run each tick.
     *
     * Incompatible with macro parameters.
     */
    runEveryTick?: boolean;
    /**
     * Optional. Whether to retain the execution context through async calls. Defaults to false.
     *
     * @warn **Waiting on Smithed Dimensions for root chunk to be functional.**
     */
    asyncContext?: boolean;
}) & ResourceClassArguments<'function'>;
export declare class _RawMCFunctionClass<PARAMS extends readonly MacroArgument[] | undefined, ENV extends readonly MacroArgument[] | undefined> extends CallableResourceClass<MCFunctionNode> {
    callback: NonNullable<MCFunctionClassArguments['callback']>;
    nested: number;
    asyncContext: NonNullable<MCFunctionClassArguments['asyncContext']>;
    protected tags: MCFunctionClassArguments['tags'];
    protected lazy: boolean;
    protected env?: ENV;
    constructor(core: SandstoneCore, name: string, args: MCFunctionClassArguments, env?: ENV);
    protected generate: () => void;
    protected addToTag: (tag: string) => void;
    __call__: (..._params: PARAMS extends undefined ? [] : PARAMS) => FinalCommandOutput;
    schedule: {
        clear: () => FinalCommandOutput;
        function: (delay: TimeArgument, type: ScheduleType) => FinalCommandOutput;
    };
    push(...contents: _RawMCFunctionClass<PARAMS, ENV>[] | [() => any]): void;
    unshift(...contents: _RawMCFunctionClass<PARAMS, ENV>[] | [() => any]): void;
    splice(start: number, removeItems: number | 'auto', ...contents: _RawMCFunctionClass<PARAMS, ENV>[] | [() => void]): void;
}
export declare const MCFunctionClass: import("sandstone/utils").MakeClassCallable<typeof _RawMCFunctionClass>;
export type MCFunctionClass<PARAMS extends readonly MacroArgument[] | undefined, ENV extends readonly MacroArgument[] | undefined> = MakeInstanceCallable<_RawMCFunctionClass<PARAMS, ENV>>;
