import { makeCallable, makeClassCallable } from 'sandstone/utils';
import { ResolveNBTPart } from '../../../variables/ResolveNBT.js';
import { ContainerNode } from '../../nodes.js';
import { CallableResourceClass, } from '../resource.js';
import { TagClass } from './tag.js';
const tags = {};
/**
 * A node representing a Minecraft function.
 */
export class MCFunctionNode extends ContainerNode {
    resource;
    contextStack;
    constructor(sandstoneCore, resource) {
        super(sandstoneCore);
        this.resource = resource;
        this.contextStack = [this];
    }
    /**
     * The currently active context.
     *
     * For example, the current context is the function body if the function is not in a loop.
     * If the function is in a loop, the current context is the loop body.
     */
    get currentContext() {
        return this.contextStack[this.contextStack.length - 1];
    }
    /**
     * Sequentially add node(s) to the end of the body of the function.
     *
     * @param node The node(s) to add.
     */
    appendNode = (node) => {
        if (Array.isArray(node)) {
            for (const _node of node) {
                this.currentContext.append(_node);
            }
        }
        else {
            this.currentContext.append(node);
        }
    };
    /**
     * Sequentially add node(s) to the beginning of the body of the function.
     *
     * @param node The node(s) to add.
     */
    prependNode = (node) => {
        if (Array.isArray(node)) {
            for (const _node of node) {
                this.currentContext.prepend(_node);
            }
        }
        else {
            this.currentContext.prepend(node);
        }
    };
    /**
     * Switch the current context to the given node.
     * Also adds the node to the body of the current context, except if addNode is False.
     *
     * @param node The node to switch to.
     * @param addNode Whether to add the node to the body of the current context.
     */
    enterContext = (node, addNode = true) => {
        if (addNode) {
            this.currentContext.append(node);
        }
        this.contextStack.push(node);
    };
    /**
     * Switch the current context to the given node, run the given function, and switch back to the original context.
     * Also adds the node to the body of the current context, except if addNode is False.
     *
     * @param node The node to switch to.
     * @param callback The function to run.
     * @param addNode Whether to add the node to the body of the current context.
     *
     * @return The previously active context.
     * @throws Error if there is no previous context.
     */
    insideContext = (node, callback, addNode = true) => {
        this.enterContext(node, addNode);
        callback();
        return this.exitContext();
    };
    /**
     * Leave the current context, and return to the previous one.
     *
     * @return The previously active context.
     * @throws Error if there is no previous context.
     */
    exitContext = () => {
        if (this.contextStack.length === 0) {
            throw new Error('No previous context to return to.');
        }
        if (this.contextStack.length === 1) {
            throw new Error('It is forbidden for a MCFunction to exit its latest context, since the MCFunction itself must be in the context stack.');
        }
        return this.contextStack.pop();
    };
    getValue = () => {
        this.sandstoneCore.currentNode = this.resource.name;
        return this.body.filter((node) => node.getValue() !== null).map((node) => node.getValue()).join('\n');
    };
}
export class _RawMCFunctionClass extends CallableResourceClass {
    callback;
    nested = 0;
    asyncContext;
    tags;
    lazy;
    env;
    constructor(core, name, args, env) {
        super(core, { packType: core.pack.dataPack(), extension: 'mcfunction' }, MCFunctionNode, core.pack.resourceToPath(name, ['functions']), {
            ...args,
            addToSandstoneCore: args.lazy ? false : args.addToSandstoneCore,
        });
        this.callback = args.callback ?? (() => { });
        if (this.nested !== 0) {
            // eslint-disable-next-line no-plusplus
            for (let i = 0; i < this.nested; i++) {
                this.node.exitContext();
            }
        }
        this.asyncContext = args.asyncContext === undefined ? false : args.asyncContext;
        this.lazy = !!args.lazy;
        this.addToSandstoneCore = !!args.addToSandstoneCore;
        this.tags = args.tags;
        if (env) {
            this.env = env;
        }
        if (args.runOnLoad) {
            if (env) {
                core.pack.loadTags.load.push(core.pack.MCFunction(`load_${this.name.split(':')[1]}`, () => {
                    /* @ts-ignore */
                    this.__call__();
                }));
            }
            else {
                core.pack.loadTags.load.push(this.name);
            }
        }
        else if (args.runEveryTick) {
            /* @ts-ignore */
            core.pack.registerTickedCommands('1t', () => this.__call__());
        }
        if (args.tags) {
            for (const tag of args.tags) {
                this.addToTag(`${tag}`);
            }
        }
        if (!args.runEveryTick && args.runEvery) {
            /* @ts-ignore */
            core.pack.registerTickedCommands(args.runEvery, () => this.__call__());
        }
        this.handleConflicts();
    }
    generate = () => {
        if (this.node.body.length > 0) {
            /*
             * Don't generate resource if the node already has commands.
             * Else, this might generate the nodes twice with fast refresh
             */
            // return
            // TODO: Fix above
        }
        this.push(this.callback.bind(undefined, makeCallable(this, this.__call__)));
    };
    addToTag = (tag) => {
        if (tags[tag]) {
            tags[tag].push(this.name);
        }
        else {
            tags[tag] = new TagClass(this.core, 'functions', tag, {
                values: [this.name], addToSandstoneCore: true, creator: 'sandstone', onConflict: 'append',
            });
        }
    };
    __call__ = (..._params) => {
        if (this.env || _params.length !== 0) {
            const args = {};
            if (_params.length !== 0) {
                for (const [i, param] of _params.entries()) {
                    // Haha funny TypeScript
                    /* @ts-ignore */
                    args[`param_${i}`] = ResolveNBTPart(param);
                    param['local'].set(this.name, `param_${i}`);
                }
                // Yeah this is cursed, but there's not really a better way to do this
                this.node.body = [];
                /* @ts-ignore */
                this.core.insideContext(this.node, () => this.callback(makeCallable(this, this.__call__.bind(undefined, ..._params)), ..._params), false);
            }
            if (this.env) {
                for (const [i, env] of this.env.entries()) {
                    // Haha funny TypeScript
                    /* @ts-ignore */
                    args[`env_${i}`] = ResolveNBTPart(env);
                    env['local'].set(this.name, `env_${i}`);
                }
            }
            return this.commands.functionCmd(this.name, 'with', this.pack.ResolveNBT(args).dataPoint);
        }
        return this.commands.functionCmd(this.name);
    };
    schedule = {
        clear: () => this.commands.schedule.clear(this.name),
        function: (delay, type) => this.commands.schedule.function(this.name, delay, type),
    };
    push(...contents) {
        if (contents[0] instanceof _RawMCFunctionClass) {
            for (const mcfunction of contents) {
                this.node.body.push(...mcfunction.node.body);
            }
        }
        else {
            this.core.enterMCFunction(this);
            this.core.insideContext(this.node, contents[0], false);
            this.core.exitMCFunction();
        }
    }
    unshift(...contents) {
        const fake = new MCFunctionClass(this.core, 'fake', {
            addToSandstoneCore: false,
            creator: 'sandstone',
            onConflict: 'ignore',
        });
        if (contents[0] instanceof _RawMCFunctionClass) {
            for (const mcfunction of contents) {
                this.node.body.unshift(...mcfunction.node.body);
            }
        }
        else {
            this.core.enterMCFunction(fake);
            this.core.insideContext(fake.node, contents[0], false);
            this.core.exitMCFunction();
            this.node.body.unshift(...fake.node.body);
        }
    }
    splice(start, removeItems, ...contents) {
        const fake = new MCFunctionClass(this.core, 'fake', {
            addToSandstoneCore: false,
            creator: 'sandstone',
            onConflict: 'ignore',
        });
        const fullBody = [];
        if (contents[0] instanceof _RawMCFunctionClass) {
            for (const mcfunction of contents) {
                fullBody.push(...mcfunction.node.body);
            }
        }
        else {
            this.core.enterMCFunction(fake);
            this.core.insideContext(fake.node, contents[0], false);
            this.core.exitMCFunction();
            fullBody.push(...fake.node.body);
        }
        this.node.body.splice(start, removeItems === 'auto' ? fullBody.length : removeItems, ...fullBody);
    }
}
export const MCFunctionClass = makeClassCallable(_RawMCFunctionClass);
