/// <reference types="node" resolution-mode="require"/>
import { ContainerNode } from '../../nodes.js';
import { ResourceClass } from '../resource.js';
import type { BLOCKS, Coordinates, ENTITY_TYPES, RootNBT, StructureNBT } from 'sandstone/arguments';
import type { ResourceClassArguments, ResourceNode, SandstoneCore } from 'sandstone/core';
import type { LiteralUnion } from 'sandstone/utils';
import type { DataPointClass, Score, StructureMirror, StructureRotation } from 'sandstone/variables';
/**
 * A node representing a Minecraft trim material.
 */
export declare class StructureNode extends ContainerNode implements ResourceNode<StructureClass> {
    resource: StructureClass;
    constructor(sandstoneCore: SandstoneCore, resource: StructureClass);
    getValue: () => Buffer | Promise<Buffer> | undefined;
}
type Block = {
    id: LiteralUnion<BLOCKS>;
    state?: {
        [key: string]: string;
    };
    nbt?: RootNBT;
};
type StructureEntry = {
    block?: Block | Block[];
    entities?: [
        {
            id: LiteralUnion<ENTITY_TYPES>;
            nbt?: RootNBT;
            offset?: [number, number, number];
        }
    ];
};
export type StructureClassArguments = {
    /**
     * Either:
     * - Path in existing resources
     * - Structure's NBT
     * - Existing Structure definition
     * - 3D Array representing the structure. Longest entry will be counted as bounds, missing or empty elements will be counted as air from origin.
     */
    structure?: string | StructureNBT | StructureClass | [[[StructureEntry]]];
} & ResourceClassArguments<'default'>;
type VariableInsertion = Score | DataPointClass<'storage'>;
export declare class StructureClass extends ResourceClass<StructureNode> {
    structureBuffer?: Promise<Buffer>;
    structureNBT?: StructureNBT;
    constructor(sandstoneCore: SandstoneCore, name: string, args: StructureClassArguments);
    structure(): StructureNBT | Promise<Buffer>;
    readBuffer(): Promise<void>;
    array(): Promise<[[[StructureEntry]]]>;
    arrayToNBT(array: [[[StructureEntry]]]): StructureNBT;
    /**
     * Places the structure.
     *
     * Usage of `VariableInsertion` or `xz` will result in a structure block being used.
     *
     * @param pos Optional. Where to place relative to the current location.
     * @param rotation Optional. Rotation to apply to the structure.
     * @param mirror Optional. Mirroring to apply to the structure. `xz` will clear all existing blocks in the region, only works for structures with defined size, & will await structures not in memory.
     * @param integrity Optional. How complete the structure will be. Must be a float between 1.0 & 0.0 (inclusive), if below 1 the structure will be randomly degraded.
     * @param seed Optional. Integer applied to the integrity random. Defaults to 0.
     */
    load(pos?: [number, number, number], rotation?: StructureRotation | VariableInsertion, mirror?: StructureMirror | '^xz' | 'xz' | VariableInsertion, integrity?: number | VariableInsertion, seed?: number | VariableInsertion): Promise<void>;
    save(pos: Coordinates<false> | [Score, Score, Score] | undefined, size: [number, number, number] | [Score, Score, Score]): void;
    toString(): string;
}
export {};
