import { toMinecraftResourceName } from 'sandstone/utils';
import { ContainerNode } from '../../nodes.js';
import { ResourceClass } from '../resource.js';
function isMCFunctionClass(v) {
    return typeof v === 'function';
}
function isTagClass(v) {
    return v instanceof TagClass;
}
function isTagObject(v) {
    return typeof v === 'object';
}
function objectToString(value) {
    if (isMCFunctionClass(value)) {
        if (value['node'].resource['env']) {
            return value['node'].sandstoneCore.pack.MCFunction(`${value.name}/_env`, () => value(), { creator: 'sandstone', onConflict: 'rename' }).name;
        }
        return value.name;
    }
    if (isTagClass(value)) {
        return value.name;
    }
    if (isTagObject(value)) {
        if (isMCFunctionClass(value.id)) {
            if (value.id['node'].resource['env']) {
                return {
                    id: value.id['node'].sandstoneCore.pack.MCFunction(`${value.id.name}/_env`, () => value.id(), { creator: 'sandstone', onConflict: 'rename' }).name,
                    required: value.required,
                };
            }
            return {
                id: value.id.name,
                required: value.required,
            };
        }
        if (isTagClass(value.id)) {
            return {
                id: value.id.name,
                required: value.required,
            };
        }
    }
    return value;
}
/**
 * A node representing a Minecraft tag.
 */
export class TagNode extends ContainerNode {
    resource;
    constructor(sandstoneCore, resource) {
        super(sandstoneCore);
        this.resource = resource;
    }
    getValue = () => JSON.stringify(this.resource.tagJSON);
}
export class TagClass extends ResourceClass {
    type;
    tagJSON;
    constructor(sandstoneCore, type, name, args) {
        super(sandstoneCore, { packType: sandstoneCore.pack.dataPack(), extension: 'json' }, TagNode, sandstoneCore.pack.resourceToPath(name, ['tags', type]), args);
        this.type = type;
        this.tagJSON = {
            replace: args.replace || false,
            values: [],
        };
        this.tagJSON.values = Array.from(args.values, objectToString);
        this.handleConflicts();
    }
    get name() {
        return `#${toMinecraftResourceName(this.path, 2)}`;
    }
    push(...resources) {
        for (const resource of resources) {
            this.tagJSON.values.push(objectToString(resource));
        }
    }
    unshift(..._resources) {
        // Done this way so the resources you're adding to the beginning of the tag stay in order.
        const resources = [];
        for (const resource of _resources) {
            resources.push(objectToString(resource));
        }
        this.tagJSON.values.push(...resources);
    }
    /** Checks whether a given resource is in this tag. */
    has(resource) {
        this.tagJSON.values.some((tagValue) => (typeof tagValue !== 'string' ? tagValue.id : tagValue) === objectToString(resource));
    }
    toString() {
        return this.name;
    }
    toJSON() {
        return toMinecraftResourceName(this.path);
    }
    /**
     * @internal
     */
    _toMinecraftCondition = () => {
        if (this.type === 'blocks' || this.type === 'entity_types') {
            return new this.pack.conditions.Tag(this.core, this.type, this.name);
        }
        throw new Error(`Cannot use a ${this.type} group tag as a condition. Only supports 'blocks' or 'entity_types'.`);
    };
}
