import { VectorClass } from 'sandstone/variables';
import { ContainerNode } from '../../nodes.js';
import { ResourceClass } from '../resource.js';
import type { SandstoneCore } from '../../sandstoneCore.js';
import type { ResourceClassArguments, ResourceNode } from '../resource.js';
type ElementFace = {
    texture: `#${string}`;
    uv?: [number, number, number, number];
    cullface?: 'north' | 'east' | 'south' | 'west' | 'up' | 'down';
    rotation?: 0 | 90 | 180 | 270;
    tintindex?: number;
};
/**
 * The maximum custom model data is the maximum integer a float can store
 * without losing precision because the predicate value is internally converted
 * to a float
 */
export declare const MAX_CUSTOM_MODEL_DATA = 16777216;
/**
 * @see https://minecraft.wiki/w/Model
 */
export type ModelData = {
    parent?: string;
    /**
     * Whether to shade the block using ambient occlusion
     * @see https://en.wikipedia.org/wiki/Ambient_occlusion
     */
    ambientocclusion?: boolean;
    textures?: {
        particle?: string;
    } & {
        [name: string]: string;
    };
    display?: Partial<Record<DisplayPosition, Partial<{
        rotation: [number, number, number];
        translation: [number, number, number];
        scale: [number, number, number];
    }>>>;
    elements?: {
        from: [number, number, number];
        to: [number, number, number];
        rotation?: {
            origin: [number, number, number];
            axis: 'x' | 'y' | 'z';
            angle: -45 | -22.5 | 0 | 22.5 | 45;
            rescale?: boolean;
        };
        shade?: boolean;
        faces?: {
            down?: ElementFace;
            up?: ElementFace;
            north?: ElementFace;
            east?: ElementFace;
            south?: ElementFace;
            west?: ElementFace;
        };
    }[];
    overrides?: {
        predicate: Partial<Record<OverridePredicate, number>>;
        model: string;
    }[];
};
declare const DisplayPositions: readonly ["thirdpirson_righthand", "thirdpirson_lefthand", "firstpirson_righthand", "firstpirson_lefthand", "gui", "head", "ground", "fixed"];
type DisplayPosition = typeof DisplayPositions[number];
declare const OverridePredicates: readonly ["angle", "blocking", "broken", "cast", "cooldown", "damage", "damaged", "lefthanded", "pull", "pulling", "charged", "firework", "throwing", "time", "custom_model_data", "trim_material"];
/**
 * - `angle` Used on compasses to determine the current angle, expressed in a decimal value of less than one.
 * - `blocking` Used on shields to determine if currently blocking. If `1`, the player is blocking.
 * - `broken` Used on Elytra to determine if broken. If `1`, the Elytra is broken.
 * - `cast` Used on fishing rods to determine if the fishing rod has been cast. If `1`, the fishing rod has been cast.
 * - `cooldown` Used on ender pearls and chorus fruit to determine the remaining cooldown, expressed in a decimal value between `0` and `1`.
 * - `damage` Used on items with durability to determine the amount of damage, expressed in a decimal value between `0` and `1`.
 * - `damaged` Used on items with durability to determine if it is damaged. If 1, the item is damaged. Note that if an item has the unbreakable tag,
 * this may be `0` while the item has a non-zero `'damage'` tag.
 * - `lefthanded` Determines the model used by left handed players. It affects the item they see in inventories, along with the item players see them holding or wearing.
 * - `pull` Determines the amount a bow or crossbow has been pulled, expressed in a decimal value of less than one.
 * - `pulling` Used on bows and crossbows to determine if the bow is being pulled. If `1`, the bow is currently being pulled.
 * - `charged` Used on crossbows to determine if they are charged with any projectile. If `1`, the crossbow is charged.
 * - `firework` Used on crossbows. If `1`, the crossbow is charged with a firework rocket.
 * - `throwing` Used on the trident to determine if the trident is ready to be thrown by the player. If `1`, the trident is ready for fire.
 * - `time` Used on clocks to determine the current time, expressed in a decimal value of less than one.
 * - `custom_model_data` Used on any item and is compared to the CustomModelData NBT, expressed in an integer value. The number is still internally converted to
 * float, causing a precision loss for some numbers above 16 million. If the value read from the item data is greater than or equal to the value used for the predicate, the predicate is positive.
 */
type OverridePredicate = typeof OverridePredicates[number];
export declare class ModelNode extends ContainerNode implements ResourceNode<ModelClass> {
    resource: ModelClass;
    constructor(sandstoneCore: SandstoneCore, resource: ModelClass);
    getValue: () => string;
}
export type ModelClassArguments = {
    /**
     * The model's path or its JSON.
     */
    model?: string | ModelData;
} & ResourceClassArguments<'default'>;
/**
 * Helper class for modifying Minecraft model data
 */
export declare class ModelClass extends ResourceClass<ModelNode> {
    type: 'block' | 'item';
    existingPath?: string;
    parent: string | null;
    ambientocclusion: boolean;
    textures: {
        [name: string]: string;
    };
    elements: ElementClass[];
    display: {
        [type in DisplayPosition]?: DisplayTransformClass;
    };
    overrides: ModelData['overrides'];
    constructor(core: SandstoneCore, type: 'block' | 'item', name: string, args: ModelClassArguments);
    toString: () => string;
    /**
     * Generates the Minecraft model JSON
     */
    toJSON(): ModelData;
    generatedItem(...layers: string[]): ModelData;
    fromJSON(data: ModelData): void;
    load(): Promise<this>;
}
type ElementArguments = NonNullable<NonNullable<ModelData['elements']>[0]>;
/**
 * Helper class for modifying model elements (cubes)
 */
export declare class ElementClass {
    to: Vector3;
    from: Vector3;
    faces: NonNullable<ElementArguments['faces']>;
    rotation: ElementArguments['rotation'];
    shade: boolean;
    /**
     * Creates a new cube without any faces
     */
    constructor(data: ElementArguments);
    hasFace(face: BlockFace): boolean;
    get hasFaces(): boolean;
    setFace(blockFace: BlockFace, face: ElementFace): void;
    deleteFace(blockFace: BlockFace): void;
    /**
     * Returns a new element with only the filtered faces
     */
    filterFaces(predicate: (blockFace: BlockFace, face: ElementFace) => boolean): ElementClass;
    /**
     * Gets the culling for a face
     */
    getFaceCull(face: BlockFace): BlockFace;
    /**
     * Returns a shallow copy of this element
     */
    clone(): ElementClass;
    /**
     * Generates the Minecraft model JSON
     */
    toJSON(): ElementArguments;
}
type DisplayTransformArguments = NonNullable<NonNullable<ModelData['display']>['head']>;
/**
 * Helper class for modifying the model transformation when displayed
 */
export declare class DisplayTransformClass {
    rotation: Vector3;
    translation: Vector3;
    scale: Vector3;
    constructor(data: DisplayTransformArguments);
    /**
     * Returns a shallow copy of this transform
     */
    clone(): DisplayTransformClass;
    toJSON(): NonNullable<DisplayTransformArguments>;
}
/**
 * All possible values of an index for a specific array as a type.
 * @example
 * ```ts
 * const array = [2, 3, 6, 9, 'f'] as const
 * type Index = 0 | 1 | 2 | 3 | 4
 * type Index = IndexOf<typeof array>
 * ```
 */
type IndexOf<A extends unknown[] | readonly unknown[]> = _IndexOf<A>;
type _IndexOf<A extends unknown[] | readonly unknown[], _S extends number[] = []> = _S['length'] extends A['length'] ? _S[number] : _IndexOf<A, [_S['length'], ..._S]>;
/**
 * An array of type T with length L.
 * This is a short way of writing homogenous array types with a fixed length
 * and is at the same time more readable.
 * @example
 * ```ts
 * type Vec4 = [number, number, number, number]
 * type Vec4 = ArrayOfLength<4, number>
 * ```
 */
type ArrayOfLength<L extends number, T> = _ArrayOfLength<L, T>;
type _ArrayOfLength<L extends number, T, _S extends T[] = []> = _S['length'] extends L ? _S : _ArrayOfLength<L, T, [..._S, T]>;
/**
 * Enumerates all numbers from 0 to exlusively A as a type.
 * @example
 * ```ts
 * type Digits = 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9
 * type Digits = EnumerateInt<10>
 * ```
 */
type EnumerateInt<A extends number> = _EnumerateInt<A>;
type _EnumerateInt<A extends number, _S extends number[] = []> = A extends _S['length'] ? _S[number] : _EnumerateInt<A, [..._S, _S['length']]>;
type Stringifyable = string | number | bigint | boolean | null | undefined;
type StringifyList<L extends Stringifyable[] | readonly Stringifyable[]> = _StringifyList<L>;
type _StringifyList<L extends Stringifyable[] | readonly Stringifyable[], R extends string[] = []> = L['length'] extends R['length'] ? R : _StringifyList<L, [...R, `${L[R['length']]}`]>;
type PrefixStrings<P extends Stringifyable, S extends Stringifyable[] | readonly Stringifyable[]> = _PrefixStrings<P, S>;
type _PrefixStrings<P extends Stringifyable, S extends Stringifyable[] | readonly Stringifyable[], R extends string[] = []> = S['length'] extends R['length'] ? R : _PrefixStrings<P, S, [...R, `${P}${S[R['length']]}`]>;
type CArray<T extends readonly number[]> = ArrayOfLength<T['length'], number>;
type CVector<T extends readonly number[]> = Vector<CArray<T>>;
type Vector3 = Vector<ArrayOfLength<3, number>>;
declare class Vector<T extends number[]> {
    readonly dimensions: T['length'];
    private readonly _values;
    /**
     * This is a strict vector constructor. In some cases it is useful because
     * Typescript knows the exact inner values, though this may lead to some type
     * incompatibility issues. Prefer to use the dedicated `vec2/3/4/N` functions.
     */
    constructor(vector: Readonly<T> | Vector<T>);
    constructor(...vector: T);
    /**
     * Gets a component of this vector
     */
    get<I extends IndexOf<T>>(idx: I): T[I];
    get(idx: number): T[number] | undefined;
    get components(): T;
    get lengthSquared(): number;
    get length(): number;
    [Symbol.iterator](): IterableIterator<number>;
    every(predicate: (value: T[IndexOf<T>], index: IndexOf<T>) => boolean): boolean;
    /**
     * Dot product
     */
    dot(...vector: CArray<T>): number;
    dot(...vector: [CVector<T>]): number;
    /**
     * Applies a map function to each component
     */
    map(mapFn: (n: T[IndexOf<T>], i: IndexOf<T>) => number): CVector<T>;
    round(): Vector<T>;
    floor(): Vector<T>;
    ceil(): Vector<T>;
    square(): Vector<T>;
    add(vector: CVector<T>): CVector<T>;
    add(...vector: CArray<T>): CVector<T>;
    add(all: number): CVector<T>;
    sub(vector: CVector<T>): CVector<T>;
    sub(...vector: CArray<T>): CVector<T>;
    sub(all: number): CVector<T>;
    mul(vector: CVector<T>): CVector<T>;
    mul(...vector: CArray<T>): CVector<T>;
    mul(all: number): CVector<T>;
    div(vector: CVector<T>): CVector<T>;
    div(...vector: CArray<T>): CVector<T>;
    div(all: number): CVector<T>;
    toCoordsAbs(): VectorClass<StringifyList<T>>;
    toCoordsRel(): VectorClass<PrefixStrings<'~', T>>;
    toCoordsLoc(): VectorClass<PrefixStrings<'^', T>>;
}
export type BlockFaceName = 'north' | 'east' | 'south' | 'west' | 'up' | 'down';
export type BlockAxisName = 'x' | 'y' | 'z';
type Unconst<A extends readonly unknown[], _S extends unknown[] = []> = A['length'] extends _S['length'] ? _S : Unconst<A, [..._S, A[_S['length']]]>;
export declare class BlockFace<Name extends BlockFaceName = BlockFaceName, Idx extends EnumerateInt<6> = EnumerateInt<6>, Normal extends Readonly<ArrayOfLength<3, -1 | 0 | 1>> = Readonly<ArrayOfLength<3, -1 | 0 | 1>>, OppositeName extends 'North' | 'East' | 'South' | 'West' | 'Up' | 'Down' = 'North' | 'East' | 'South' | 'West' | 'Up' | 'Down'> {
    readonly name: Name;
    readonly id: Idx;
    static readonly North: BlockFace<"north", 0, readonly [0, 0, -1], "South">;
    static readonly East: BlockFace<"east", 1, readonly [1, 0, 0], "West">;
    static readonly South: BlockFace<"south", 2, readonly [0, 0, 1], "North">;
    static readonly West: BlockFace<"west", 3, readonly [-1, 0, 0], "East">;
    static readonly Up: BlockFace<"up", 4, readonly [0, 1, 0], "Down">;
    static readonly Down: BlockFace<"down", 5, readonly [0, -1, 0], "Up">;
    static fromName(name: 'north'): (typeof BlockFace)['North'];
    static fromName(name: 'east'): (typeof BlockFace)['East'];
    static fromName(name: 'south'): (typeof BlockFace)['South'];
    static fromName(name: 'west'): (typeof BlockFace)['West'];
    static fromName(name: 'up'): (typeof BlockFace)['Up'];
    static fromName(name: 'down'): (typeof BlockFace)['Down'];
    static fromName(name: BlockFaceName): BlockFace;
    static fromName(name: string): BlockFace | null;
    static [Symbol.iterator](): IterableIterator<BlockFace<"north", 0, readonly [0, 0, -1], "South"> | BlockFace<"east", 1, readonly [1, 0, 0], "West"> | BlockFace<"south", 2, readonly [0, 0, 1], "North"> | BlockFace<"west", 3, readonly [-1, 0, 0], "East"> | BlockFace<"up", 4, readonly [0, 1, 0], "Down"> | BlockFace<"down", 5, readonly [0, -1, 0], "Up">>;
    static get filter(): {
        <S extends BlockFace<"north", 0, readonly [0, 0, -1], "South"> | BlockFace<"east", 1, readonly [1, 0, 0], "West"> | BlockFace<"south", 2, readonly [0, 0, 1], "North"> | BlockFace<"west", 3, readonly [-1, 0, 0], "East"> | BlockFace<"up", 4, readonly [0, 1, 0], "Down"> | BlockFace<"down", 5, readonly [0, -1, 0], "Up">>(predicate: (value: BlockFace<"north", 0, readonly [0, 0, -1], "South"> | BlockFace<"east", 1, readonly [1, 0, 0], "West"> | BlockFace<"south", 2, readonly [0, 0, 1], "North"> | BlockFace<"west", 3, readonly [-1, 0, 0], "East"> | BlockFace<"up", 4, readonly [0, 1, 0], "Down"> | BlockFace<"down", 5, readonly [0, -1, 0], "Up">, index: number, array: readonly (BlockFace<"north", 0, readonly [0, 0, -1], "South"> | BlockFace<"east", 1, readonly [1, 0, 0], "West"> | BlockFace<"south", 2, readonly [0, 0, 1], "North"> | BlockFace<"west", 3, readonly [-1, 0, 0], "East"> | BlockFace<"up", 4, readonly [0, 1, 0], "Down"> | BlockFace<"down", 5, readonly [0, -1, 0], "Up">)[]) => value is S, thisArg?: any): S[];
        (predicate: (value: BlockFace<"north", 0, readonly [0, 0, -1], "South"> | BlockFace<"east", 1, readonly [1, 0, 0], "West"> | BlockFace<"south", 2, readonly [0, 0, 1], "North"> | BlockFace<"west", 3, readonly [-1, 0, 0], "East"> | BlockFace<"up", 4, readonly [0, 1, 0], "Down"> | BlockFace<"down", 5, readonly [0, -1, 0], "Up">, index: number, array: readonly (BlockFace<"north", 0, readonly [0, 0, -1], "South"> | BlockFace<"east", 1, readonly [1, 0, 0], "West"> | BlockFace<"south", 2, readonly [0, 0, 1], "North"> | BlockFace<"west", 3, readonly [-1, 0, 0], "East"> | BlockFace<"up", 4, readonly [0, 1, 0], "Down"> | BlockFace<"down", 5, readonly [0, -1, 0], "Up">)[]) => unknown, thisArg?: any): (BlockFace<"north", 0, readonly [0, 0, -1], "South"> | BlockFace<"east", 1, readonly [1, 0, 0], "West"> | BlockFace<"south", 2, readonly [0, 0, 1], "North"> | BlockFace<"west", 3, readonly [-1, 0, 0], "East"> | BlockFace<"up", 4, readonly [0, 1, 0], "Down"> | BlockFace<"down", 5, readonly [0, -1, 0], "Up">)[];
    };
    static forEach(callback: (face: BlockFace, id: number) => void): void;
    readonly opposite: (typeof BlockFace)[OppositeName];
    readonly normal: Vector<Unconst<Normal>>;
    readonly isSide: Name extends 'north' | 'east' | 'south' | 'west' ? true : false;
    readonly axis: Name extends 'east' | 'west' ? 'x' : Name extends 'up' | 'down' ? 'y' : 'z';
    private constructor();
    toString(): Name;
    toJSON(): Name;
}
export {};
