/* eslint-disable @typescript-eslint/no-non-null-assertion */
/* eslint-disable max-len */
import { VectorClass } from 'sandstone/variables';
import { ContainerNode } from '../../nodes.js';
import { ResourceClass } from '../resource.js';
const assert = (condition, errorMessage) => {
    if (!condition) {
        throw new Error(errorMessage);
    }
};
/**
 * The maximum custom model data is the maximum integer a float can store
 * without losing precision because the predicate value is internally converted
 * to a float
 */
export const MAX_CUSTOM_MODEL_DATA = 16777216;
const DisplayPositions = ['thirdpirson_righthand', 'thirdpirson_lefthand', 'firstpirson_righthand', 'firstpirson_lefthand', 'gui', 'head', 'ground', 'fixed'];
const OverridePredicates = ['angle', 'blocking', 'broken', 'cast', 'cooldown', 'damage', 'damaged', 'lefthanded', 'pull', 'pulling', 'charged', 'firework', 'throwing', 'time', 'custom_model_data', 'trim_material'];
export class ModelNode extends ContainerNode {
    resource;
    constructor(sandstoneCore, resource) {
        super(sandstoneCore);
        this.resource = resource;
    }
    getValue = () => JSON.stringify(this.resource.toJSON());
}
/**
 * Helper class for modifying Minecraft model data
 */
export class ModelClass extends ResourceClass {
    type;
    existingPath;
    parent = null;
    ambientocclusion = true;
    textures = {};
    elements = [];
    display = {};
    // TODO: Helper methods
    overrides = [];
    constructor(core, type, name, args) {
        super(core, { packType: core.pack.resourcePack() }, ModelNode, core.pack.resourceToPath(name, ['models', type]), args);
        this.type = type;
        const data = args.model;
        if (data) {
            if (typeof data === 'string') {
                this.existingPath = data;
            }
            else {
                this.fromJSON(data);
            }
        }
        this.handleConflicts();
    }
    toString = () => `${this.path[0]}:${this.path.slice(2)}`;
    /**
     * Generates the Minecraft model JSON
     */
    toJSON() {
        const obj = {};
        if (this.parent) {
            obj.parent = this.parent;
        }
        if (Object.keys(this.textures).length > 0) {
            obj.textures = this.textures;
        }
        if (!this.ambientocclusion) {
            obj.ambientocclusion = false;
        }
        if (this.elements && this.elements.length > 0) {
            obj.elements = this.elements.map((e) => e.toJSON());
        }
        if (Object.keys(this.display).length > 0) {
            obj.display = {};
            const entries = Object.entries(this.display);
            entries.forEach(([type, t]) => {
                obj.display[type] = t.toJSON();
            });
        }
        if (this.overrides && this.overrides.length > 0) {
            obj.overrides = this.overrides;
        }
        return obj;
    }
    generatedItem(...layers) {
        const textures = {};
        // eslint-disable-next-line no-return-assign
        layers.forEach((texture, idx) => (textures[`layer${idx}`] = texture));
        return {
            parent: 'minecraft:item/generated',
        };
    }
    fromJSON(data) {
        if (data.parent) {
            this.parent = data.parent;
        }
        if (data.textures) {
            this.textures = data.textures;
        }
        if (data.display) {
            Object.values(DisplayPositions)
                .filter((t) => t in data.display)
                .forEach((t) => {
                this.display[t] = (new DisplayTransformClass(data.display[t]));
            });
        }
        if (typeof data.ambientocclusion === 'boolean') {
            this.ambientocclusion = data.ambientocclusion;
        }
        if (data.elements) {
            this.elements = data.elements.map((element) => (new ElementClass(element)));
        }
        if (data.overrides) {
            this.overrides = data.overrides;
        }
    }
    async load() {
        this.fromJSON(JSON.parse(await this.core.getExistingResource(this)));
        return this;
    }
}
/**
 * Helper class for modifying model elements (cubes)
 */
export class ElementClass {
    to;
    from;
    faces = {};
    rotation;
    shade = true;
    /**
     * Creates a new cube without any faces
     */
    constructor(data) {
        this.to = new Vector(...data.to);
        this.from = new Vector(...data.from);
        if (data.faces)
            this.faces = data.faces;
        if (data.rotation)
            this.rotation = data.rotation;
        if (data.shade)
            this.shade = data.shade;
    }
    hasFace(face) {
        if (this.hasFaces) {
            return face.name in this.faces;
        }
        return false;
    }
    get hasFaces() {
        return Object.keys(this.faces).length > 0;
    }
    setFace(blockFace, face) {
        this.faces[blockFace.name] = face;
    }
    deleteFace(blockFace) {
        delete this.faces[blockFace.name];
    }
    /**
     * Returns a new element with only the filtered faces
     */
    filterFaces(predicate) {
        const clone = this.clone();
        const faces = Object.entries(this.faces);
        faces.filter(([fname, face]) => predicate(BlockFace.fromName(fname), face)).forEach(([fname, face]) => { clone.faces[fname] = face; });
        return clone;
    }
    /**
     * Gets the culling for a face
     */
    getFaceCull(face) {
        const _face = this.faces[face.name];
        return _face?.cullface ? BlockFace.fromName(_face.cullface) : face;
    }
    /**
     * Returns a shallow copy of this element
     */
    clone() {
        const clone = new ElementClass(this.toJSON());
        clone.rotation = this.rotation;
        clone.shade = this.shade;
        return clone;
    }
    /**
     * Generates the Minecraft model JSON
     */
    toJSON() {
        const obj = {
            from: this.from.components,
            to: this.to.components,
        };
        if (this.hasFaces) {
            obj.faces = this.faces;
        }
        if (this.rotation) {
            obj.rotation = this.rotation;
        }
        if (!this.shade) {
            obj.shade = false;
        }
        return obj;
    }
}
/**
 * Helper class for modifying the model transformation when displayed
 */
export class DisplayTransformClass {
    rotation = Vector3Props.ZERO;
    translation = Vector3Props.ZERO;
    scale = Vector3Props.ONE;
    constructor(data) {
        if (data.rotation)
            this.rotation = new Vector(data.rotation);
        if (data.translation)
            this.translation = new Vector(data.translation);
        if (data.scale)
            this.scale = new Vector(data.scale);
    }
    /**
     * Returns a shallow copy of this transform
     */
    clone() {
        const clone = new DisplayTransformClass(this.toJSON());
        clone.rotation = this.rotation;
        clone.translation = this.translation;
        clone.scale = this.scale;
        return clone;
    }
    toJSON() {
        const obj = {};
        if (this.rotation.every((n) => n === 0))
            obj.rotation = this.rotation.components;
        if (this.translation.every((n) => n === 0))
            obj.translation = this.translation.components;
        if (this.scale.every((n) => n === 1))
            obj.scale = this.scale.components;
        return obj;
    }
}
class Vector {
    dimensions;
    _values;
    constructor(...args) {
        if (typeof args[0] === 'number') {
            this._values = args.slice();
        }
        else {
            const arg = args[0];
            if (arg instanceof Vector) {
                this._values = arg._values;
            }
            else {
                this._values = arg;
            }
        }
        this.dimensions = args.length;
        Object.freeze(this._values);
        Object.freeze(this);
    }
    get(idx) {
        return this._values[idx];
    }
    get components() {
        return Object.freeze(this._values.slice());
    }
    get lengthSquared() {
        return this._values.reduce((a, b) => a + b * b, 0);
    }
    get length() {
        return Math.sqrt(this.lengthSquared);
    }
    [Symbol.iterator]() {
        return this._values[Symbol.iterator]();
    }
    every(predicate) {
        return this._values.every(predicate);
    }
    dot(...args) {
        // @ts-ignore TS is broken here
        if (typeof args[0] === 'number') {
            // @ts-ignore TS is broken here
            assert(args.length === this.dimensions, 'Vectors do not have a matching length');
            // @ts-ignore TS is broken here
            return this._values.map((n, i) => n * args[i]).reduce((a, b) => a + b);
        }
        // @ts-ignore TS is broken here
        const other = args[0];
        assert(other.dimensions === this.dimensions, 'Vectors do not have a matching length');
        // @ts-ignore TS is broken here
        return this._values.map((n, i) => n * other._values[i]).reduce((a, b) => a + b);
    }
    /**
     * Applies a map function to each component
     */
    map(mapFn) {
        // @ts-ignore TS can't handle the types
        return new Vector(this._values.map(mapFn));
    }
    round() {
        // @ts-ignore TS broken https://github.com/Microsoft/TypeScript/issues/29112
        return this.map(Math.round);
    }
    floor() {
        // @ts-ignore TS broken https://github.com/Microsoft/TypeScript/issues/29112
        return this.map(Math.floor);
    }
    ceil() {
        // @ts-ignore TS broken https://github.com/Microsoft/TypeScript/issues/29112
        return this.map(Math.ceil);
    }
    square() {
        // @ts-ignore TS broken https://github.com/Microsoft/TypeScript/issues/29112
        return this.map((n) => n * n);
    }
    add(...args) {
        // @ts-ignore TS is broken here
        if (typeof args[0] === 'number') {
            // @ts-ignore TS is broken here
            assert(args.length === this.dimensions, 'Vectors do not have a matching length');
            // @ts-ignore TS can't handle the types
            return new Vector(Object.freeze(this._values.map((n, i) => n + args[i])));
        }
        // @ts-ignore TS is broken here
        const other = args[0];
        if (other instanceof Vector) {
            assert(other.dimensions === this.dimensions, 'Vectors do not have a matching length');
            return new Vector(Object.freeze(this._values.map((n, i) => n + other._values[i])));
        }
        return new Vector(Object.freeze(this._values.map((n) => n + other)));
    }
    sub(...args) {
        // @ts-ignore TS is broken here
        if (typeof args[0] === 'number') {
            // @ts-ignore TS is broken here
            assert(args.length === this.dimensions, 'Vectors do not have a matching length');
            // @ts-ignore TS can't handle the types
            return new Vector(Object.freeze(this._values.map((n, i) => n - args[i])));
        }
        // @ts-ignore TS is broken here
        const other = args[0];
        if (other instanceof Vector) {
            assert(other.dimensions === this.dimensions, 'Vectors do not have a matching length');
            return new Vector(Object.freeze(this._values.map((n, i) => n - other._values[i])));
        }
        return new Vector(Object.freeze(this._values.map((n) => n - other)));
    }
    mul(...args) {
        // @ts-ignore TS is broken here
        if (typeof args[0] === 'number') {
            // @ts-ignore TS is broken here
            assert(args.length === this.dimensions, 'Vectors do not have a matching length');
            // @ts-ignore TS can't handle the types
            return new Vector(Object.freeze(this._values.map((n, i) => n * args[i])));
        }
        // @ts-ignore TS is broken here
        const other = args[0];
        if (other instanceof Vector) {
            assert(other.dimensions === this.dimensions, 'Vectors do not have a matching length');
            return new Vector(Object.freeze(this._values.map((n, i) => n * other._values[i])));
        }
        return new Vector(Object.freeze(this._values.map((n) => n * other)));
    }
    div(...args) {
        // @ts-ignore TS is broken here
        if (typeof args[0] === 'number') {
            // @ts-ignore TS is broken here
            assert(args.length === this.dimensions, 'Vectors do not have a matching length');
            // @ts-ignore TS can't handle the types
            return new Vector(Object.freeze(this._values.map((n, i) => n / args[i])));
        }
        // @ts-ignore TS is broken here
        const other = args[0];
        if (other instanceof Vector) {
            assert(other.dimensions === this.dimensions, 'Vectors do not have a matching length');
            return new Vector(Object.freeze(this._values.map((n, i) => n / other._values[i])));
        }
        return new Vector(Object.freeze(this._values.map((n) => n / other)));
    }
    toCoordsAbs() {
        return new VectorClass(this._values.map((n) => `${n}`));
    }
    toCoordsRel() {
        return new VectorClass(this._values.map((n) => `~${n}`));
    }
    toCoordsLoc() {
        return new VectorClass(this._values.map((n) => `^${n}`));
    }
}
const Vector2Props = {
    ZERO: new Vector(0, 0),
    ONE: new Vector(1, 1),
};
function vec2(x, y) {
    if (typeof x === 'number') {
        return new Vector(x, y);
    }
    return new Vector(x);
}
const Vector3Props = {
    ZERO: new Vector(0, 0, 0),
    ONE: new Vector(1, 1, 1),
};
function vec3(x, y, z) {
    if (typeof x === 'number') {
        return new Vector(x, y, z);
    }
    return new Vector(x);
}
const Vector4Props = {
    ZERO: new Vector(0, 0, 0, 0),
    ONE: new Vector(1, 1, 1, 1),
};
function vec4(x, y, z, w) {
    if (typeof x === 'number') {
        return new Vector(x, y, z, w);
    }
    return new Vector(x);
}
function vecN(...args) {
    if (typeof args[0] === 'number') {
        return new Vector(args);
    }
    return new Vector(args[0]);
}
export class BlockFace {
    name;
    id;
    static North = new BlockFace('north', 0, [0, 0, -1], 'South');
    static East = new BlockFace('east', 1, [1, 0, 0], 'West');
    static South = new BlockFace('south', 2, [0, 0, 1], 'North');
    static West = new BlockFace('west', 3, [-1, 0, 0], 'East');
    static Up = new BlockFace('up', 4, [0, 1, 0], 'Down');
    static Down = new BlockFace('down', 5, [0, -1, 0], 'Up');
    static fromName(name) {
        switch (name) {
            case 'north': return this.North;
            case 'east': return this.East;
            case 'south': return this.South;
            case 'west': return this.West;
            case 'up': return this.Up;
            case 'down': return this.Down;
            default: return null;
        }
    }
    static [Symbol.iterator]() {
        return BLOCK_FACES[Symbol.iterator]();
    }
    static get filter() {
        return BLOCK_FACES.filter;
    }
    static forEach(callback) {
        BLOCK_FACES.forEach((face, id) => callback(face, id));
    }
    opposite;
    /* @ts-ignore */
    normal;
    isSide;
    axis;
    constructor(name, id, normal, opposite) {
        this.name = name;
        this.id = id;
        /* @ts-ignore */
        this.normal = new Vector(normal);
        if (normal[0] !== 0)
            this.axis = 'x';
        else if (normal[1] !== 0)
            this.axis = 'y';
        else
            this.axis = 'z';
        this.isSide = (this.axis !== 'y');
        // is set correctly later
        this.opposite = opposite;
    }
    toString() {
        return this.name;
    }
    toJSON() {
        return this.name;
    }
}
const BLOCK_FACES = Object.freeze([
    BlockFace.North,
    BlockFace.East,
    BlockFace.South,
    BlockFace.West,
    BlockFace.Up,
    BlockFace.Down,
]);
BlockFace.forEach((face) => {
    const opp = face.opposite;
    face.opposite = BlockFace[opp];
});
