/* eslint-disable operator-linebreak */
import fs from 'fs-extra';
import path from 'path';
import { DataPackDependencies, ResourcePackDependencies } from '../pack/dependencies.js';
import { MCMetaCache } from './mcmeta.js';
import { SmithedDependencyClass } from './resources/dependency.js';
import { SmithedDependencyCache } from './smithed.js';
export class SandstoneCore {
    pack;
    /** All Resources */
    resourceNodes;
    mcfunctionStack;
    awaitNodes;
    currentNode = '';
    mcmetaCache = new MCMetaCache();
    smithed = new SmithedDependencyCache(this);
    dependencies = [];
    constructor(pack) {
        this.pack = pack;
        this.resourceNodes = new Set();
        this.mcfunctionStack = [];
        this.awaitNodes = new Set();
        // ESM is funny
        for (const method of Object.getOwnPropertyNames(Object.getPrototypeOf(this))) {
            /* @ts-ignore */
            if (method !== 'constructor' && typeof this[method] === 'function' && typeof this[method].bind === 'function') {
                /* @ts-ignore */
                this[method] = this[method].bind(this);
            }
        }
    }
    /**
     * The current MCFunction.
     */
    get currentMCFunction() {
        return this.mcfunctionStack[this.mcfunctionStack.length - 1];
    }
    getCurrentMCFunctionOrThrow = () => {
        const { currentMCFunction } = this;
        if (!currentMCFunction) {
            throw new Error('This operation is invalid when outside a MCFunction.');
        }
        return currentMCFunction;
    };
    insideContext = (...args) => this.getCurrentMCFunctionOrThrow().insideContext(...args);
    /**
     * Create a new MCFunction with the given name, and switch the currently active MCFunction to it.
     * @param mcfunction The MCFunction to switch to.
     * @return The newly created and active MCFunction.
     */
    enterMCFunction = (mcfunction) => {
        /*
         * We cannot simply call mcfunction.node, because .node is protected to avoid polluting the autocompleted API.
         * However, TypeScript gives us a backdoor using this dynamic call, in a fully type-safe way.
         */
        // eslint-disable-next-line prefer-destructuring, dot-notation
        const node = mcfunction['node'];
        this.mcfunctionStack.push(node);
        return node;
    };
    /**
     * Leave the current MCFunction, and return to the previous one.
     * @return The previously active MCFunction.
     */
    exitMCFunction = () => this.mcfunctionStack.pop();
    async getExistingResource(pathOrResource, encoding = 'utf-8') {
        if (typeof pathOrResource === 'string') {
            if (encoding === false) {
                return fs.readFile(pathOrResource);
            }
            return fs.readFile(pathOrResource, encoding);
        }
        const _path = pathOrResource.path;
        if (_path[0] === 'minecraft') {
            const type = pathOrResource.packType.resourceSubFolder;
            return this.mcmetaCache.get(type, `${type}/${_path.join('/')}${pathOrResource.fileExtension ? `.${pathOrResource.fileExtension}` : ''}`, (encoding === 'utf-8'));
        }
        // eslint-disable-next-line max-len
        const fullPath = path.join(process.env.WORKING_DIR, `resources/${path.join(pathOrResource.packType.type, ..._path)}${pathOrResource.fileExtension ? `.${pathOrResource.fileExtension}` : ''}`);
        if (pathOrResource.fileEncoding === false) {
            return fs.readFile(fullPath);
        }
        return fs.readFile(fullPath, pathOrResource.fileEncoding);
    }
    async getVanillaResource(relativePath, text = true, type = 'server') {
        return this.mcmetaCache.get(type === 'server' ? 'data' : 'assets', relativePath, text);
    }
    /**
     * Add a dependency for a Smithed Library
     */
    depend(dependency, version = 'latest') {
        const adding = this.dependencies.length === 0 ? false : Promise.allSettled(this.dependencies);
        this.dependencies.push((async () => {
            if (!this.smithed.loaded) {
                await this.smithed.load();
            }
            if (adding)
                await adding;
            if (!this.smithed.has(dependency)) {
                const _depend = await this.smithed.get(dependency, version);
                if (!this.pack.packTypes.has('datapack-dependencies')) {
                    this.pack.packTypes.set('datapack-dependencies', new DataPackDependencies());
                }
                const dependencies = [new SmithedDependencyClass(this, dependency, _depend, 'server')];
                if (_depend.resourcepack) {
                    if (!this.pack.packTypes.has('resourcepack-dependencies')) {
                        this.pack.packTypes.set('resourcepack-dependencies', new ResourcePackDependencies());
                    }
                    dependencies.push(new SmithedDependencyClass(this, dependency, _depend, 'client'));
                }
                return dependencies;
            }
            return true;
        })());
    }
    generateResources = (opts) => {
        const originalResources = new Set(this.resourceNodes);
        // First, generate all the resources.
        for (const { resource } of this.resourceNodes) {
            resource['generate']();
        }
        // Then, transform all the nodes with the given visitors.
        for (const visitor of opts.visitors) {
            visitor.onStart();
            for (const node of this.resourceNodes) {
                visitor.visit(node);
            }
            visitor.onEnd();
        }
        // Since visitors may change the resources, swap back to the previous ones.
        const finalResources = this.resourceNodes;
        this.resourceNodes = originalResources;
        return finalResources;
    };
    save = async (cliOptions, opts) => {
        if (!this.smithed.loaded) {
            await this.smithed.load();
        }
        await Promise.allSettled(this.dependencies);
        await this.mcmetaCache.save();
        const resources = this.generateResources(opts);
        for await (const node of resources) {
            const { packType, fileExtension } = node.resource;
            const _path = [packType.type, ...node.resource.path];
            if (packType.resourceSubFolder) {
                _path.splice(1, 0, packType.resourceSubFolder);
            }
            const resourcePath = path.join(..._path);
            const value = node.getValue();
            if (cliOptions.verbose) {
                console.log(`Path: ${resourcePath}.${fileExtension}\n\n` +
                    `${typeof value === 'string' ? value : '<Buffer>'}`);
            }
            if (!cliOptions.dry) {
                await cliOptions.fileHandler(`${resourcePath}.${fileExtension}`, value);
            }
        }
    };
}
