/* eslint-disable @typescript-eslint/no-non-null-assertion */
import AdmZip from 'adm-zip';
import fs from 'fs-extra';
import path from 'path';
import { fetch, safeWrite } from 'sandstone/utils';
export class SmithedDependencyCache {
    core;
    constructor(core) {
        this.core = core;
    }
    baseURL = 'https://api.smithed.dev/v2/';
    path = path.join(process.env.WORKING_DIR, '../', 'resources', 'cache', 'smithed');
    manifest = path.join(this.path, '..', '..', 'smithed.json');
    lockFile = path.join(this.path, '..', 'lock-smithed.json');
    dependencies = new Map();
    loaded = false;
    async load() {
        this.loaded = true;
        if (!(await fs.pathExists(this.manifest))) {
            return;
        }
        const manifest = JSON.parse(await fs.readFile(this.manifest, 'utf-8'));
        for (const [id, version] of Object.entries(manifest)) {
            this.core.depend(id, version);
        }
    }
    async save() {
        if (this.dependencies.size !== 0) {
            const dependencies = {};
            for (const [dependency, { version }] of this.dependencies.entries()) {
                dependencies[dependency] = version;
            }
            await safeWrite(this.manifest, JSON.stringify(dependencies));
            const lock = {};
            for (const [name, { version, date, resourcepack }] of this.dependencies.entries()) {
                lock[name] = {
                    version,
                    date,
                    resourcepack: !(!resourcepack),
                };
            }
            await safeWrite(this.lockFile, JSON.stringify(lock));
        }
    }
    async get(dependency, version) {
        if (!this.loaded) {
            await this.load();
        }
        const existing = this.dependencies.get(dependency);
        if (existing && existing.version === version) {
            existing.added = true;
            return existing;
        }
        let result;
        const lockFile = await (async () => {
            try {
                return JSON.parse(await fs.readFile(this.lockFile, 'utf-8'));
            }
            catch (e) {
                return false;
            }
        })();
        let download = false;
        if (!lockFile || !lockFile[dependency] || lockFile[dependency].version !== version) {
            download = true;
        }
        else {
            const date = Date.now();
            if (version === 'latest' && ((lockFile[dependency].date - date) / 36e5) > 1) {
                lockFile[dependency].date = date;
                download = true;
            }
            else {
                result = this.dependencies.set(dependency, {
                    version,
                    date: lockFile[dependency].date,
                    datapack: await fs.readFile(path.join(this.path, dependency, 'datapack.zip')),
                    resourcepack: lockFile[dependency].resourcepack ? await fs.readFile(path.join(this.path, dependency, 'resourcepack.zip')) : false,
                    added: false,
                }).get(dependency);
            }
        }
        if (download) {
            let url = `${this.baseURL}download?pack=${dependency}`;
            let hasResourcePack = false;
            const pack = JSON.parse(await (await fetch(`${this.baseURL}packs/${dependency}`)).text());
            if (version === 'latest') {
                const { versions } = pack;
                const ver = versions[versions.length - 1];
                if (ver.downloads.resourcepack) {
                    hasResourcePack = true;
                }
            }
            else {
                const ver = pack.versions.find((_ver) => _ver.name === version);
                if (ver.downloads.resourcepack) {
                    hasResourcePack = true;
                }
                url += `@${version}`;
            }
            const files = (new AdmZip(await (await fetch(url)).buffer())).getEntries();
            const datapack = await new Promise((res) => {
                files[0].getDataAsync((data) => res(data));
            });
            await safeWrite(path.join(this.path, dependency, 'datapack.zip'), datapack);
            let resourcepack = false;
            if (hasResourcePack) {
                resourcepack = await new Promise((res) => {
                    files[1].getDataAsync((data) => res(data));
                });
                await safeWrite(path.join(this.path, dependency, 'resourcepack.zip'), datapack);
            }
            result = this.dependencies.set(dependency, {
                version,
                date: version === 'latest' ? Date.now() : 0,
                datapack,
                resourcepack,
                added: true,
            }).get(dependency);
            await this.save();
        }
        return result;
    }
    has(dependency) {
        const depend = this.dependencies.get(dependency);
        return depend && depend.added;
    }
}
