/// <reference types="node" resolution-mode="require"/>
import { SandstoneCommands } from 'sandstone/commands';
import { AdvancementClass, AtlasClass, BlockStateClass, DamageTypeClass, FontClass, ItemModifierClass, LanguageClass, LootTableClass, MacroLiteral, MCFunctionClass, ModelClass, PlainTextClass, PredicateClass, RecipeClass, SandstoneCore, SoundEventClass, TagClass, TextureClass, TrimMaterialClass, TrimPatternClass } from 'sandstone/core';
import { CustomResourceClass } from 'sandstone/core/resources/custom';
import { Flow } from 'sandstone/flow';
import { DataArray, DataClass, DataIndexMap, DataPointClass, LoopArgument, ObjectiveClass, SelectorClass, TargetlessDataClass, TargetlessDataPointClass } from 'sandstone/variables';
import { ResolveNBTClass } from 'sandstone/variables/ResolveNBT';
import { Score } from 'sandstone/variables/Score';
import { SleepClass } from 'sandstone/variables/Sleep';
import { UUIDClass } from 'sandstone/variables/UUID';
import { PackType } from './packType.js';
import type { AdvancementJSON, AtlasDefinition, BlockStateDefinition, BlockStateType, Coordinates, DamageTypeJSON, FontProvider, ItemModifierJSON, JSONTextComponent, LootTableJSON, NBTObject, OBJECTIVE_CRITERION, PredicateJSON, RecipeJSON, REGISTRIES, RootNBT, SingleEntityArgument, SOUND_TYPES, TagValuesJSON, TEXTURE_TYPES, TimeArgument, TrimMaterialJSON, TrimPatternJSON } from 'sandstone/arguments';
import type { StoreType } from 'sandstone/commands';
import type { _RawMCFunctionClass, AdvancementClassArguments, AtlasClassArguments, BlockStateArguments, DamageTypeClassArguments, DataPointPickClass, FontArguments, ItemModifierClassArguments, LanguageArguments, LootTableClassArguments, MacroArgument, MCFunctionClassArguments, ModelClassArguments, Node, PlainTextArguments, PredicateClassArguments, RecipeClassArguments, SoundEventArguments, TagClassArguments, TextureArguments, TrimMaterialClassArguments, TrimPatternClassArguments } from 'sandstone/core';
import type { LiteralUnion, MakeInstanceCallable } from 'sandstone/utils';
import type { DATA_PATH, DATA_TYPES, SelectorCreator } from 'sandstone/variables';
import type { UUIDinNumber, UUIDinScore, UUIDOptions } from 'sandstone/variables/UUID';
import type { handlerReadFile, handlerWriteFile } from './packType.js';
export type ResourcePath = string[];
type MCFunctionArgs = Exclude<Partial<MCFunctionClassArguments>, MCFunctionClassArguments['callback']>;
export declare class DataPack extends PackType {
    readonly packMcmeta: any;
    constructor(archiveOutput: boolean, options: {
        packFormat: number;
        description: JSONTextComponent;
        features?: string[];
        filter?: {
            namespace?: string;
            path?: string;
        }[];
    });
    handleOutput: (type: 'output' | 'client' | 'server', readFile: handlerReadFile, writeFile: handlerWriteFile) => Promise<void>;
}
export declare class ResourcePack extends PackType {
    readonly packMcmeta: any;
    constructor(options: {
        packFormat: number;
        description: JSONTextComponent;
        features?: string[];
        filter?: {
            namespace?: string;
            path?: string;
        }[];
    });
    handleOutput: (type: 'output' | 'client' | 'server', readFile: handlerReadFile, writeFile: handlerWriteFile) => Promise<void>;
}
export declare class SandstonePack {
    defaultNamespace: string;
    packUid: string;
    readonly core: SandstoneCore;
    packTypes: Map<string, PackType>;
    packOptions: any;
    dataPack(): DataPack;
    resourcePack(): ResourcePack;
    dependencies: Map<string, boolean>;
    readonly flow: Flow;
    readonly commands: SandstoneCommands<false>;
    readonly Macro: SandstoneCommands<true> & ((strings: TemplateStringsArray, ...macros: (string | number | MacroArgument)[]) => MacroLiteral);
    readonly conditions: {
        Advancement: typeof import("../flow/conditions/resources/advancement.js").AdvancementConditionNode;
        Predicate: typeof import("../flow/conditions/resources/predicate.js").PredicateConditionNode;
        Tag: typeof import("../flow/conditions/resources/tag.js").TagConditionNode;
        TrimMaterial: typeof import("../flow/conditions/resources/trimMaterial.js").TrimMaterialConditionNode;
        TrimPattern: typeof import("../flow/conditions/resources/trimPattern.js").TrimPatternConditionNode;
        DataPoint: typeof import("../flow/conditions/variables/dataPoint.js").DataPointExistsConditionNode;
        DataPointEquals: typeof import("../flow/conditions/variables/dataPoint.js").DataPointEqualsConditionNode;
        Label: typeof import("../flow/conditions/variables/label.js").LabelConditionNode;
        Score: typeof import("sandstone/flow").ScoreConditionNode;
        Selector: typeof import("../flow/conditions/variables/selector.js").SelectorConditionNode;
        UUID: typeof import("../flow/conditions/variables/uuid.js").UUIDConditionNode;
    };
    objectives: Set<ObjectiveClass>;
    anonymousScoreId: number;
    anonymousDataId: number;
    constants: Set<number>;
    tickedLoops: Record<string, MakeInstanceCallable<_RawMCFunctionClass<undefined, undefined>>>;
    loadTags: {
        preLoad: TagClass<'functions'>;
        load: TagClass<'functions'>;
        postLoad: TagClass<'functions'>;
    };
    constructor(defaultNamespace: string, packUid: string);
    setupLantern: () => void;
    resourceToPath: (name: string, resourceFolders?: string[]) => ResourcePath;
    /** UTILS */
    get _(): Flow;
    registerNewObjective: (objective: ObjectiveClass) => void;
    /**
     * Register a new numeric constant.
     */
    registerNewConstant(amount: number): void;
    Objective: {
        /**
         * Create a new objective. Defaults to `dummy` if unspecified.
         * @param name The name of the objective
         */
        create: (name: string, criteria?: LiteralUnion<OBJECTIVE_CRITERION>, display?: JSONTextComponent, alreadyExists?: boolean) => ObjectiveClass;
        /** Get an existing objective. */
        get: (name: string) => ObjectiveClass;
    };
    __rootObjective?: ObjectiveClass;
    get rootObjective(): ObjectiveClass;
    Variable: ((
    /**
     * Creates a dynamic numeric variable, represented by an anonymous & unique score.
     *
     * @param initialValue The initial value of the variable. If left unspecified,
     * or if `undefined`, then the score will not be initialized.
     *
     * @param name A name that can be useful for debugging.
     */
    (initialValue?: number | Score | undefined, name?: string) => Score) & (
    /**
     * Creates a dynamic numeric variable, represented by an anonymous & unique score.
     *
     * @param nbt The NBT value to set the Variable to.
     *
     * @param scale The scale to multiply the value by. Defaults to 1.
     *
     * @param name A name that can be useful for debugging.
     */
    (nbt: DataPointClass, scale?: number, name?: string) => Score));
    get flowVariable(): Score;
    /**
     * Creates a new label
     * @param label Label/tag name
     * @param description Label description (optional)
     */
    Label: (label: string, description?: string) => MakeInstanceCallable<import("sandstone/variables")._RawLabelClass>;
    /**
     * Creates a new Data Class Instance
     * @param type Source type
     * @param target Optional. Target for data
     * @param path Optional. Path for data
     */
    Data<T extends DATA_TYPES>(type: T): TargetlessDataClass<T>;
    Data<T extends 'entity'>(type: T, target: SingleEntityArgument<false>): DataClass<T>;
    Data<T extends 'storage'>(type: T, target: string): DataClass<T>;
    Data<T extends 'block'>(type: T, target: Coordinates<false>): DataClass<T>;
    Data<T extends DATA_TYPES>(type: T, target: undefined, path: DATA_PATH | DATA_PATH[]): TargetlessDataPointClass<T>;
    Data<T extends 'entity'>(type: T, target: SingleEntityArgument<false>, path: DATA_PATH | DATA_PATH[]): DataPointClass<T>;
    Data<T extends 'storage'>(type: T, target: string, path: DATA_PATH | DATA_PATH[]): DataPointClass<T>;
    Data<T extends 'block'>(type: T, target: Coordinates<false>, path: DATA_PATH | DATA_PATH[]): DataPointClass<T>;
    __rootStorage?: DataPointClass;
    rootStorage(): DataPointClass<any>;
    DataVariable: ((
    /**
     * Creates a dynamic data variable, represented by an anonymous & unique Data Point.
     *
     * @param initialValue Optional. The initial value of the variable. If left unspecified,
     * or if `undefined`, then the Data Point will not be initialized.
     *
     * @param name Optional. A name that can be useful for debugging.
     */
    (initialValue?: NBTObject | DataPointClass<any> | DataPointPickClass, name?: string) => DataPointClass<'storage'>) & (
    /**
     * Creates a dynamic data variable, represented by an anonymous & unique Data Point.
     *
     * @param score The Score to set the data variable to.
     *
     * @param storeType Optional. The number type you want to store. Defaults to int.
     *
     * @param scale Optional. The scale to multiply the value by. Defaults to 1.
     *
     * @param name Optional. A name that can be useful for debugging.
     */
    (score: Score, storeType?: StoreType, scale?: number, name?: string) => DataPointClass<'storage'>));
    DataIndexMap<INITIAL extends RootNBT | Record<string, DataPointClass | DataPointPickClass>>(initialize: INITIAL, dataPoint?: DataPointClass<'storage'>): DataIndexMap<INITIAL>;
    DataArray<INITIAL extends readonly NBTObject[] | readonly [string, DataPointClass | DataPointPickClass]>(initialize: INITIAL, dataPoint?: DataPointClass<'storage'>): DataArray<INITIAL>;
    getTempStorage: (name: string) => DataPointClass<"storage">;
    /**
     * Resolve NBT containing Data Points & Score's.
     * @param nbt NBT with Data Points & Scores initialized with ResolveNBTPart.
     * @param dataPoint Optional. Data Point to resolve to.
     */
    ResolveNBT: (nbt: NBTObject, dataPoint?: DataPointClass<'storage'>) => ResolveNBTClass;
    Selector: SelectorCreator<false>;
    /** Initializes with a static UUID. (you can use `import { randomUUID } from 'sandstone/utils'` to generate one) */
    UUID(source: string | UUIDinNumber, options?: UUIDOptions): UUIDClass<'known'>;
    /** Initializes with `Score`'s. */
    UUID(source: UUIDinScore, options?: UUIDOptions): UUIDClass<'scores'>;
    /** Initializes with a Data Point (to a UUID int array). */
    UUID(source: DataPointClass, options?: UUIDOptions): UUIDClass<'data'>;
    /** Initializes with a Selector. */
    UUID(source: SelectorClass<true, boolean>, options?: UUIDOptions): UUIDClass<'selector'>;
    MCFunction(name: string, options?: MCFunctionArgs): MCFunctionClass<undefined, undefined>;
    MCFunction(name: string, callback: (loop: MCFunctionClass<undefined, undefined>) => void, options?: MCFunctionArgs): MCFunctionClass<undefined, undefined>;
    MCFunction<PARAMS extends readonly MacroArgument[]>(name: string, callback: (loop: MCFunctionClass<PARAMS, undefined>, ...params: PARAMS) => void, options?: MCFunctionArgs): MCFunctionClass<PARAMS, undefined>;
    MCFunction<ENV extends readonly MacroArgument[]>(name: string, environmentVariables: ENV, callback: (loop: MCFunctionClass<undefined, ENV>) => void, options?: MCFunctionArgs): MCFunctionClass<undefined, ENV>;
    MCFunction<PARAMS extends readonly MacroArgument[], ENV extends readonly MacroArgument[]>(name: string, environmentVariables: ENV, callback: (loop: MCFunctionClass<PARAMS, ENV>, ...params: PARAMS) => void, options?: MCFunctionArgs): MCFunctionClass<PARAMS, ENV>;
    appendNode: (node: Node) => void;
    __initMCFunction?: MCFunctionClass<undefined, undefined>;
    get initMCFunction(): MCFunctionClass<undefined, undefined>;
    /**
     * Register commands that will be ticked at the rate you specify
     */
    registerTickedCommands(runEvery: TimeArgument, callback: () => void): void;
    sleep: (delay: TimeArgument) => PromiseLike<SleepClass>;
    Loop: () => LoopArgument;
    __customResourceTypes: string[];
    registerCustomResource(type: string): void;
    get makeCustomResource(): [SandstoneCore, typeof CustomResourceClass];
    /** Creates a resource in the datapack, path must include file extension. */
    RawResource(path: string, contents: string | Buffer | Promise<Buffer>): CustomResourceClass;
    /** Creates a resource in the given pack, path must include file extension. */
    RawResource(pack: PackType, path: string, contents: string | Buffer | Promise<Buffer>): CustomResourceClass;
    Advancement: <T extends string>(name: string, advancement: AdvancementJSON<T>, options?: Partial<AdvancementClassArguments>) => AdvancementClass<string>;
    DamageType: (name: string, damageType: DamageTypeJSON, options?: Partial<DamageTypeClassArguments>) => DamageTypeClass;
    ItemModifier: (name: string, itemModifier: ItemModifierJSON, options?: Partial<ItemModifierClassArguments>) => ItemModifierClass;
    LootTable: (name: string, lootTable: LootTableJSON, options?: Partial<LootTableClassArguments>) => LootTableClass;
    Predicate: (name: string, predicate: PredicateJSON, options?: Partial<PredicateClassArguments>) => PredicateClass;
    Recipe: (name: string, recipe: RecipeJSON, options?: Partial<RecipeClassArguments>) => RecipeClass;
    /** @ts-ignore */
    Tag: <T extends LiteralUnion<REGISTRIES>>(type: T, name: string, values: TagValuesJSON<T>, options?: Partial<TagClassArguments<T>> | undefined) => TagClass<T>;
    TrimMaterial: (name: string, trimMaterial: TrimMaterialJSON, options?: Partial<TrimMaterialClassArguments>) => TrimMaterialClass;
    TrimPattern: (name: string, trimPattern: TrimPatternJSON, options?: Partial<TrimPatternClassArguments>) => TrimPatternClass;
    Atlas: (name: string, atlas: AtlasDefinition, options?: Partial<AtlasClassArguments>) => AtlasClass;
    BlockState<Type extends BlockStateType>(type: Type, name: string, blockState: BlockStateDefinition<Type>, options?: Partial<BlockStateArguments<Type>>): BlockStateClass<Type>;
    Font: (name: string, providers: FontProvider[], options?: Partial<FontArguments>) => FontClass;
    Language: (name: string, language: LanguageArguments['language'], options?: Partial<LanguageArguments>) => LanguageClass;
    Model: (type: 'block' | 'item', name: string, model: ModelClassArguments['model'], options?: Partial<ModelClassArguments>) => ModelClass;
    SoundEvent<Type extends SOUND_TYPES>(type: Type, name: string, sound: SoundEventArguments['sound'], options?: Partial<SoundEventArguments>): SoundEventClass<Type>;
    PlainText: (name: string, text: PlainTextArguments['text'], options?: Partial<PlainTextArguments>) => PlainTextClass;
    Texture<Type extends TEXTURE_TYPES>(type: Type, name: string, texture: TextureArguments<Type>['texture'], options?: Partial<TextureArguments<Type>>): TextureClass<Type>;
    save: (cliOptions: {
        fileHandler: (relativePath: string, content: any) => Promise<void>;
        dry: boolean;
        verbose: boolean;
    }) => Promise<Map<string, PackType>>;
}
export {};
