import { DataPointPickClass, MacroArgument } from '../core/Macro.js';
import { nbtStringifier } from './nbt/NBTs.js';
import { Score } from './Score.js';
export function NBTpathToString(path) {
    let result = '';
    for (const p of path) {
        if (typeof p === 'string') {
            if (result.length > 0) {
                result += '.';
            }
            result += p;
        }
        else {
            result += nbtStringifier(p);
        }
    }
    return result;
}
export class TargetlessDataClass {
    sandstonePack;
    type;
    constructor(sandstonePack, type) {
        this.sandstonePack = sandstonePack;
        this.type = type;
    }
    target = (target) => new DataClass(this.sandstonePack, this.type, target);
    select = (...path) => new TargetlessDataPointClass(this.sandstonePack, this.type, path);
}
export class TargetlessDataPointClass {
    sandstonePack;
    type;
    path;
    constructor(sandstonePack, type, path) {
        this.sandstonePack = sandstonePack;
        this.type = type;
        this.path = NBTpathToString(path);
    }
    target = (target) => new DataPointClass(this.sandstonePack, this.type, target, [this.path]);
    select = (...path) => new TargetlessDataPointClass(this.sandstonePack, this.type, [this.path, ...path]);
}
export class DataClass {
    sandstonePack;
    type;
    currentTarget;
    constructor(sandstonePack, type, target) {
        this.sandstonePack = sandstonePack;
        this.type = type;
        this.currentTarget = target;
    }
    /**
     * Merge the given NBT to the current target.
     */
    merge = (value) => {
        this.sandstonePack.commands.data.merge[this.type](this.currentTarget, value);
    };
    target = (target) => new DataClass(this.sandstonePack, this.type, target);
    select = (...path) => new DataPointClass(this.sandstonePack, this.type, this.currentTarget, path);
    toString = () => this.currentTarget;
    toJSON = this.toString;
}
export class DataPointClass extends MacroArgument {
    sandstonePack;
    type;
    path;
    currentTarget;
    constructor(sandstonePack, type, target, path) {
        super(sandstonePack.core);
        this.sandstonePack = sandstonePack;
        this.type = type;
        this.path = NBTpathToString(path);
        this.currentTarget = target;
    }
    target = (target) => new DataPointClass(this.sandstonePack, this.type, target, [this.path]);
    select = (...path) => new DataPointClass(this.sandstonePack, this.type, this.currentTarget, [this.path, ...path]);
    modify = (cb, value) => {
        const data = cb(this.sandstonePack.commands.data.modify[this.type](this.currentTarget, this.path));
        // The value is another Data Point
        if (value instanceof DataPointClass) {
            data.from[value.type](value.currentTarget, value.path);
            return this;
        }
        if (value instanceof DataPointPickClass) {
            this.set(value._toDataPoint());
            return this;
        }
        // The value is a NBT
        data.value(value);
        return this;
    };
    string = (cb, value, start, end) => {
        const data = cb(this.sandstonePack.commands.data.modify[this.type](this.currentTarget, this.path));
        if (!end)
            data.string[value.type](value.currentTarget, value.path, start);
        else if (start === 0) {
            data.from[value.type](value.currentTarget, value.path);
        }
        else {
            data.string[value.type](value.currentTarget, value.path, start, end);
        }
    };
    executeStore = (storeType, scale) => this.sandstonePack.commands.execute.store.result[this.type](this.currentTarget, this.path, storeType, scale);
    set = (value, storeType, scale = 1) => {
        if (value instanceof StringDataPointClass) {
            if (value.sliceBounds[1])
                this.string((data) => data.set, value, value.sliceBounds[0], value.sliceBounds[1]);
            else
                this.string((data) => data.set, value, value.sliceBounds[0]);
            return this;
        }
        if (value instanceof Score) {
            this.executeStore(storeType || 'int', scale).run.scoreboard.players.get(value.target, value.objective);
            return this;
        }
        this.modify((data) => data.set, value);
        return this;
    };
    /**
     * Set the data point to the given NBT.
     */
    merge = (value) => this.modify((data) => data.merge, value);
    /**
     * Append the given NBT to the current data point.
     */
    append = (value) => this.modify((data) => data.append, value);
    /**
     * Prepend the given NBT to the current data point.
     */
    prepend = (value) => this.modify((data) => data.prepend, value);
    /**
     * Insert the given NBT to the given index of the current data point.
     */
    insert = (value, index) => this.modify((data) => data.insert(index), value);
    /**
     * Remove the current NBT value.
     */
    remove = () => this.sandstonePack.commands.data.remove[this.type](this.currentTarget, this.path);
    /**
     * Extracts a section of a string available for setting to another path, without modifying the original path.
     *
     * Use slice(0) to get a type-safe string data point; it compiles to a regular `from` argument.
     *
     * @param start Index of first character to include at the start of the string.
     * @param end Optional. Index of the first character to exclude at the end of the string.
     */
    slice = (start, end) => new StringDataPointClass(this.sandstonePack, this.type, this.currentTarget, this.path, start, end);
    _toMinecraftCondition = () => new this.sandstonePack.conditions.DataPoint(this.sandstonePack.core, this);
    equals = (value) => new this.sandstonePack.conditions.DataPointEquals(this.sandstonePack.core, this, value);
    /**
     * @internal
     */
    _toChatComponent = () => ({
        nbt: this.path,
        [this.type]: this.currentTarget,
    });
}
export class StringDataPointClass extends DataPointClass {
    sandstonePack;
    type;
    sliceBounds;
    constructor(sandstonePack, type, target, path, start, end) {
        super(sandstonePack, type, target, [path]);
        this.sandstonePack = sandstonePack;
        this.type = type;
        this.sliceBounds = [start];
        if (end)
            this.sliceBounds.push(end);
    }
}
