import { MacroArgument } from '../core/Macro.js';
import type { FormattingTags, JSONTextComponent, MultipleEntitiesArgument, ObjectiveArgument, Range } from 'sandstone/arguments';
import type { SandstoneCommands } from 'sandstone/commands';
import type { NotNode } from 'sandstone/flow';
import type { ConditionClass } from 'sandstone/variables';
import type { SandstonePack } from '../pack/index.js';
import type { ComponentClass } from './abstractClasses.js';
import type { DataPointClass } from './Data.js';
import type { ObjectiveClass } from './Objective.js';
type PlayersTarget = number | MultipleEntitiesArgument<false>;
export type ScoreDisplay = {
    name?: JSONTextComponent;
    numberformat?: 'blank' | ['styled', FormattingTags] | ['fixed', JSONTextComponent];
};
export declare class Score extends MacroArgument implements ConditionClass, ComponentClass {
    sandstonePack: SandstonePack;
    target: MultipleEntitiesArgument<false>;
    display: ScoreDisplay | undefined;
    objective: ObjectiveClass;
    commands: SandstoneCommands<false>;
    constructor(sandstonePack: SandstonePack, target: MultipleEntitiesArgument<false>, display: ScoreDisplay | undefined, objective: ObjectiveClass);
    toString(): string;
    toJSON(): {
        readonly type: "minecraft:score";
        readonly target: {
            readonly type: "minecraft:fixed";
            readonly name: string;
        };
        readonly score: string;
        readonly scale: 1;
    };
    /**
     * @internal
     */
    _toChatComponent(): JSONTextComponent;
    /**
     * @internal
     */
    _toMinecraftCondition: () => NotNode;
    private unaryOperation;
    private binaryOperation;
    setDisplay: (display: ScoreDisplay) => this;
    /** INLINE OPERATORS */
    /**
     * Reset the entity's score.
     */
    reset: () => void;
    /**
     * Set the current entity's score to other entities's scores.
     *
     * @param targets The targets to get the scores from
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    set(targets: PlayersTarget, objective?: ObjectiveArgument): Score;
    /**
     * Set the current entity's score to the given value, or to the other target's score.
     *
     * @param amountOrTargetScore A value, or the target's score.
     */
    set(amountOrTargetScore: number | Score): Score;
    /**
     * Set the current entity's score to the given NBT value, with the given scale.
     *
     * @param nbt The Data Point to set the score to.
     *
     * @param scale The scale factor.
     */
    set(nbt: DataPointClass, scale?: number): Score;
    '=': {
        (targets: PlayersTarget, objective?: ObjectiveArgument): Score;
        (amountOrTargetScore: number | Score): Score;
        (nbt: DataPointClass, scale?: number): Score;
    };
    /**
     * Adds other entities's scores to the current entity's score.
     *
     * @param targets The targets to add the scores from.
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    add(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
    /**
     * Adds the given amount, or the other target's score, to the current entity's score.
     *
     * @param amountOrTargetScore The amount to add, or the target to add the scores from.
     */
    add(amountOrTargetScore: number | Score): Score;
    '+=': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
        (amountOrTargetScore: number | Score): Score;
    };
    /**
     * Subtract other target's scores from the current entity's score.
     *
     * @param targets The targets to get the scores from.
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    remove(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
    /**
     * Subtract the given amount, or the other target's score, from the current entity's score.
     *
     * @param targetScore The amount to subtract, or the target to get the score from.
     */
    remove(amountOrTargetScore: number | Score): Score;
    '-=': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
        (amountOrTargetScore: number | Score): Score;
    };
    /**
     * Adds one to the current entity's score.
     */
    increase(): Score;
    get '++'(): Score;
    /**
     * Subtracts one from the current entity's score.
     */
    decrease(): Score;
    get '--'(): Score;
    /**
     * Multiply the current entity's score by other entities's scores.
     *
     * @param targets The targets to get the scores from
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    multiply(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
    /**
     * Multiply the current entity's score by the given value, or other target's scores.
     *
     * @param amountOrTargetScore The value, or the target to get the scores from.
     */
    multiply(amountOrTargetScore: number | Score): Score;
    '*=': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
        (amountOrTargetScore: number | Score): Score;
    };
    /**
     * Divide the current entity's score by other entities's scores.
     *
     * @param targets The targets to get the scores from
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    divide(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
    /**
     * Divide the current entity's score by the given value, or the other target's scores.
     *
     * @param amountOrTargetScore The value, or the target to get the scores from
     */
    divide(amountOrTargetScore: number | Score): Score;
    '/=': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
        (amountOrTargetScore: number | Score): Score;
    };
    /**
     * Get the remainder of the division of the current entity's score by other entities's scores.
     *
     * @param targets The targets to get the scores from
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    modulo(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
    /**
     * Divide the current entity's score by other target's scores.
     *
     * @param amountOrTargetScore The value, or target's score to modulo the current score with.
     */
    modulo(amountOrTargetScore: number | Score): Score;
    '%=': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
        (amountOrTargetScore: number | Score): Score;
    };
    /**
     * Swap the current score with the other targets' scores.
     *
     * @param targets The targets to swap the scores with
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    swap(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): void;
    /**
     * Swap the current entity's score with the other target's scores.
     *
     * @param targetScore The target to swap the scores with
     */
    swap(targetScore: Score): void;
    '><': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): void;
        (targetScore: Score): void;
    };
    /** EFFECT-FREE OPERATORS */
    /**
     * Returns a new anonymous score, equal to the sum of the current score and the given targets' score.
     *
     * @param targets The targets to add the scores from
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    plus(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
    /**
     * Returns a new anonymous score, equal to the sum of the current score and the given amount or target' score.
     *
     * @param amountOrTargetScore The value, or the target to add the score from.
     */
    plus(amountOrTargetScore: number | Score): Score;
    '+': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
        (amountOrTargetScore: number | Score): Score;
    };
    /**
     * Returns a new anonymous score, equal to the difference between the current score and the given targets' score.
     *
     * @param targets The targets to subtract the scores from
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    minus(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
    /**
     * Returns a new anonymous score, equal to the difference between the current score and the given amount or target' score.
     *
     * @param amountOrTargetScore The amount to subtract, or the target to subtract the score from.
     */
    minus(amountOrTargetScore: number | Score): Score;
    '-': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
        (amountOrTargetScore: number | Score): Score;
    };
    /**
     * Returns a new anonymous score, equal to the product of the current score and the given targets' score.
     *
     * @param targets The targets to multiply the scores from
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    multipliedBy(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
    /**
     * Returns a new anonymous score, equal to the product of the current score and the given amount or target's score.
     *
     * @param amountOrTargetScore The amount, or the target to multiply the scores from
     */
    multipliedBy(amountOrTargetScore: number | Score): Score;
    '*': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
        (amountOrTargetScore: number | Score): Score;
    };
    /**
     * Returns a new anonymous score, equal to the division of the current score and the given targets' score.
     *
     * @param targets The targets to divide the scores by
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    dividedBy(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
    /**
     * Returns a new anonymous score, equal to the division of the current score and the given amount or target's score.
     *
     * @param amountOrTargetScore The amount, or target's score to divide the current score by.
     */
    dividedBy(amountOrTargetScore: number | Score): Score;
    '/': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
        (amountOrTargetScore: number | Score): Score;
    };
    /**
     * Returns a new anonymous score, equal to the modulo of the current score and the given targets' score.
     *
     * @param targets The targets to modulo the scores by
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    moduloBy(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
    /**
     * Returns a new anonymous score, equal to the modulo of the current score and the given value, or target's score.
     *
     * @param amountOrTargetScore The amount, or target's score to modulo the current score by.
     */
    moduloBy(amountOrTargetScore: number | Score): Score;
    '%': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): Score;
        (amountOrTargetScore: number | Score): Score;
    };
    /** COMPARISONS OPERATORS */
    private comparison;
    /**
     * Check if the current score is strictly greater than the given score.
     *
     * @param targets The target to compare the current score against.
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    greaterThan(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): ConditionClass;
    /**
     * Check if the current score is strictly greater than the given amount or score.
     *
     * @param amountOrTargetScore The amount or score to compare the current score against.
     */
    greaterThan(amountOrTargetScore: number | Score): ConditionClass;
    '>': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): ConditionClass;
        (amountOrTargetScore: number | Score): ConditionClass;
    };
    /**
     * Check if the current score is greater or equal than the given score.
     *
     * @param targets The target to compare the current score against.
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    greaterOrEqualThan(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): ConditionClass;
    /**
     * Check if the current score is greater or equal than the given amount or score.
     *
     * @param amountOrTargetScore The amount or score compare the current score against.
     */
    greaterOrEqualThan(amountOrTargetScore: number | Score): ConditionClass;
    '>=': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): ConditionClass;
        (amountOrTargetScore: number | Score): ConditionClass;
    };
    /**
     * Check if the current score is strictly lower than the given score.
     *
     * @param targets The target to compare the current score against.
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    lessThan(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): ConditionClass;
    /**
     * Check if the current score is strictly lower than the given amount or score.
     *
     * @param amountOrTargetScore The amount or score to compare the current score against.
     */
    lessThan(amountOrTargetScore: number | Score): ConditionClass;
    '<': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): ConditionClass;
        (amountOrTargetScore: number | Score): ConditionClass;
    };
    /**
     * Check if the current score is lower or equal than the given score.
     *
     * @param targets The target to compare the current score against.
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    lessOrEqualThan(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): ConditionClass;
    /**
     * Check if the current score is lower or equal than the given amount or score.
     *
     * @param amountOrTargetScore The amount or score target to compare the current score against.
     */
    lessOrEqualThan(amountOrTargetScore: number | Score): ConditionClass;
    '<=': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): ConditionClass;
        (amountOrTargetScore: number | Score): ConditionClass;
    };
    /**
     * Check if the current score is equal to than the given score.
     *
     * @param targets The target to compare the current score against.
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    equalTo(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): ConditionClass;
    /**
     * Check if the current score is equal to the given amount or score.
     *
     * @param amountOrTargetScore The amount or score to compare the current score against.
     */
    equalTo(amountOrTargetScore: number | Score): ConditionClass;
    '==': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): ConditionClass;
        (amountOrTargetScore: number | Score): ConditionClass;
    };
    /**
     * Check if the current score is not equal to than the given score.
     *
     * @param targets The target to compare the current score against.
     *
     * @param objective The related objective. If not specified, default to the same objective as the current target.
     */
    notEqualTo(targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): NotNode;
    /**
     * Check if the current score is not equal to the given amount or score.
     *
     * @param amountOrTargetScore The amount or score to compare the current score against.
     */
    notEqualTo(amountOrTargetScore: number | Score): NotNode;
    '!=': {
        (targets: MultipleEntitiesArgument<false>, objective?: ObjectiveArgument): NotNode;
        (amountOrTargetScore: number | Score): NotNode;
    };
    /**
     * Check if the current score matches a certain range.
     *
     * @param range The range to compare the current score against.
     */
    matches: (range: Range<false>) => {
        _toMinecraftCondition: () => import("sandstone/flow").ScoreConditionNode;
    };
    match: (minimum: number, maximum: number, callback: (num: number) => void) => import("../core/index.js").MCFunctionClass<undefined, this[]>;
}
export {};
