import type { ENTITY_TYPES, GAMEMODES, JSONTextComponent, Range, RootNBT } from 'sandstone/arguments';
import type { PredicateClass } from 'sandstone/core';
import type { SandstonePack } from 'sandstone/pack';
import type { Macroable } from '../core/Macro.js';
import type { LiteralUnion } from '../utils.js';
import type { ConditionTextComponentClass, SelectorPickClass } from './abstractClasses.js';
import type { LabelClass } from './Label.js';
import type { NotNBT } from './nbt/NBTs.js';
type ScoreArgument<MACRO extends boolean> = Record<string, Range<MACRO>>;
type AdvancementsArgumentValue = boolean | [string, boolean] | [string, boolean][];
type AdvancementsArgument = Record<string, AdvancementsArgumentValue | Record<string, AdvancementsArgumentValue>>;
type TypeOrArray<T> = T | T[];
/**
 * If MustBeSingle is false, then anything is allowed.
 * If it is true, then you must provide limit=1 or limit=0.
 */
export type SelectorProperties<MustBeSingle extends boolean, MustBePlayer extends boolean, MACRO extends boolean> = {
    /**
     * Filter target selection based on their Euclidean distances from some point,
     * searching for the target's feet (a point at the bottom of the center of their hitbox).
     *
     * If the positional arguments `x`, `y` and `z` are left undefined,
     * radius is calculated relative to the position of the command's execution.
     *
     * Only unsigned values are allowed.
     */
    distance?: Macroable<Range<MACRO>, MACRO>;
    /** Filter target selection based on their scores in the specified objectives. */
    scores?: Macroable<ScoreArgument<MACRO>, MACRO>;
    /**
     * Filter target selection to those who are on a given team.
     *
     * Allowed values are:
     * - `teamName`, to filter those who are in the given team.
     * - `!teamName`, to filter those who are not in the given team.
     * - `false`, to filter those who are teamless (in no team).
     * - `true`, ot filter those who have at least one team.
     */
    team?: Macroable<string | boolean, MACRO>;
    /**
     * Filter target selection to those that have the given tag(s).
     *
     * Multiple values are allowed, when passed as an array.
     *
     * @example
     *
     * Selector(`@e`, { tag: 'alive' }) => `@e[tag=alive]`
     *
     * Selector(`@e`, { tag: ['alive', 'winner'] }) => `@e[tag=alive, tag=winner]`
     *
     */
    tag?: Macroable<TypeOrArray<string | LabelClass>, MACRO>;
    /** Filter target selection to all those with a given name. This cannot be a JSON text compound. */
    name?: Macroable<string, MACRO>;
    /**
     * Specify the targets selection priority.
     *
     * - `nearest`: Sort by increasing distance. (Default for `@p`)
     * - `furthest`: Sort by decreasing distance.
     * - `random`: Sort randomly. (Default for `@r`)
     * - `arbitrary`: Do not sort. (Default for `@e`, `@a`)
     */
    sort?: 'nearest' | 'furthest' | 'random' | 'abitrary';
    /** Filter target selection based on their experience levels. This naturally filters out all non-player targets. */
    level?: Macroable<Range<MACRO>, MACRO>;
    /** Filter target selection to those who are in the specified game mode. */
    gamemode?: Macroable<LiteralUnion<GAMEMODES | `!${GAMEMODES}`> | `!${GAMEMODES}`[], MACRO>;
    /**
     * Filter target selection based on their vertical rotation, measured in degrees.
     *
     * Values range from `-90` (straight up) to `0` (at the horizon) to `+90` (straight down).
     */
    x_rotation?: Macroable<Range<MACRO>, MACRO>;
    /**
     * Filter target selection based on their rotation in the horizontal XZ-plane,
     * measured clockwise in degrees from due south (or the positive Z direction).
     *
     * Values vary:
     * - from `-180` (facing due north, the -Z direction)
     * - to `-90` (facing due east, the +X direction)
     * - to `0` (facing due south, the +Z direction)
     * - to `+90` (facing due west, the -X direction)
     * - to `+180` (facing due north again).
     */
    y_rotation?: Macroable<Range<MACRO>, MACRO>;
    /** Select all targets that match the specified advancement and value. */
    advancements?: Macroable<AdvancementsArgument, MACRO>;
    /** Select all targets that match the specified predicate. */
    predicate?: Macroable<string | PredicateClass | (PredicateClass | string)[], MACRO>;
    /** Select all targets that have the specified NBT. */
    nbt?: Macroable<RootNBT | NotNBT | (RootNBT | NotNBT)[], MACRO>;
    /**
     * Define a position on the X-axis in the world the selector starts at,
     * for use with the `distance` argument or the volume arguments, `dx`, `dy` and `dz`.
     */
    x?: Macroable<number, MACRO>;
    /**
     * Define a position on the Y-axis in the world the selector starts at,
     * for use with the `distance` argument or the volume arguments, `dx`, `dy` and `dz`.
     */
    y?: Macroable<number, MACRO>;
    /**
     * Define a position on the Z-axis in the world the selector starts at,
     * for use with the `distance` argument or the volume arguments, `dx`, `dy` and `dz`.
     */
    z?: Macroable<number, MACRO>;
    /**
     * Filter target selection based on their x-difference, from some point,
     * as measured from the closest corner of the entities' hitboxes
     */
    dx?: Macroable<number, MACRO>;
    /**
     * Filter target selection based on their y-difference, from some point,
     * as measured from the closest corner of the entities' hitboxes
     */
    dy?: Macroable<number, MACRO>;
    /**
     * Filter target selection based on their z-difference, from some point,
     * as measured from the closest corner of the entities' hitboxes
     */
    dz?: Macroable<number, MACRO>;
} & (MustBeSingle extends true ? {
    limit: Macroable<0 | 1, MACRO>;
} : {
    limit?: Macroable<number, MACRO>;
}) & (MustBePlayer extends true ? {
    /**
     * Filter target selection to those of a specific entity type.
     *
     * Multiple values are allowed, when passed as an array.
     *
     * @example
     *
     * Selector(`@e`, { type: 'minecraft:cow' }) => `@e[type=!minecraft:cow]`
     *
     * Selector(`@e`, { type: ['!minecraft:cow', '!minecraft:skeleton'] }) => `@e[type=!minecraft:cow, type=!minecraft:skeleton]`
     */
    type: 'minecraft:player';
} : {
    /**
     * Filter target selection to those of a specific entity type.
     *
     * Multiple values are allowed, when passed as an array.
     *
     * @example
     *
     * Selector(`@e`, { type: 'minecraft:cow' }) => `@e[type=!minecraft:cow]`
     *
     * Selector(`@e`, { type: ['!minecraft:cow', '!minecraft:skeleton'] }) => `@e[type=!minecraft:cow, type=!minecraft:skeleton]`
     */
    type?: Macroable<LiteralUnion<ENTITY_TYPES> | LiteralUnion<ENTITY_TYPES>[], MACRO>;
});
export declare class SelectorClass<MACRO extends boolean, IsSingle extends boolean = false, IsPlayer extends boolean = false> implements ConditionTextComponentClass, SelectorPickClass<IsSingle, IsPlayer> {
    protected sandstonePack: SandstonePack;
    target: '@s' | '@p' | '@a' | '@e' | '@r';
    arguments: SelectorProperties<IsSingle, IsPlayer, MACRO>;
    constructor(sandstonePack: SandstonePack, target: '@s' | '@p' | '@a' | '@e' | '@r', selectorArguments?: SelectorProperties<IsSingle, IsPlayer, MACRO>);
    /**
     * List all scores of this entity.
     */
    listScores: () => void;
    toString(): string;
    /**
     * @internal
     */
    _toMinecraftCondition: () => import("../flow/conditions/variables/selector.js").SelectorConditionNode;
    /**
     * @internal
     */
    _toChatComponent(): JSONTextComponent;
    /**
     * @internal
     */
    _toSelector(): SelectorClass<MACRO, IsSingle, IsPlayer>;
    protected toJSON(): string;
}
export type AnySelectorProperties<MACRO extends boolean> = SelectorProperties<false, false, MACRO>;
export type SingleSelectorProperties<MACRO extends boolean> = SelectorProperties<true, false, MACRO>;
export type SinglePlayerSelectorProperties<MACRO extends boolean> = SelectorProperties<true, true, MACRO>;
export type SelectorCreator<MACRO extends boolean> = (((target: '@p' | '@r', selectorArguments?: Omit<AnySelectorProperties<MACRO>, 'limit' | 'type'>) => SelectorClass<MACRO, true, true>) & ((target: '@s', selectorArguments?: Omit<AnySelectorProperties<MACRO>, 'limit'>) => SelectorClass<MACRO, true, true>) & ((target: '@a', selectorArguments: Omit<SingleSelectorProperties<MACRO>, 'type'>) => SelectorClass<MACRO, true, true>) & ((target: '@a', selectorArguments?: Omit<AnySelectorProperties<MACRO>, 'type'>) => SelectorClass<MACRO, false, true>) & ((target: '@e', selectorArguments: SinglePlayerSelectorProperties<MACRO>) => SelectorClass<MACRO, true, true>) & ((target: '@e', selectorArguments: SingleSelectorProperties<MACRO>) => SelectorClass<MACRO, true, false>) & ((target: '@e', selectorArguments?: AnySelectorProperties<MACRO>) => SelectorClass<MACRO, false, false>));
export {};
