/* eslint-disable no-lone-blocks */
/* eslint-disable no-plusplus */
/* eslint-disable max-len */
import { SelectorPickClass } from './abstractClasses.js';
import { NBTIntArray } from './nbt/index.js';
import { ResolveNBTPart } from './ResolveNBT.js';
// Conversion methods ported from https://github.com/AjaxGb/mc-uuid-converter/blob/master/convert.js
const UUIDData = new DataView((new Uint8Array(16)).buffer);
const UUID_GROUP_SIZES = [8, 4, 4, 4, 12];
export class UUIDClass {
    core;
    primarySource;
    known;
    scores;
    data;
    selector;
    /**
     *
     * @param _source A primary source for the UUID. Either a UUID string, a UUID integer array, 4 scores set to the UUID, a selector, or a Data Point set to the UUID.
     */
    constructor(core, _source, options) {
        this.core = core;
        let source = _source;
        if (typeof source === 'string') {
            source = this.stringToArray(_source);
        }
        if (Array.isArray(source)) {
            // Thanks Typescript
            if (typeof source[0] === 'number') {
                this.primarySource = 'known';
                this.known = source;
            }
            else {
                this.primarySource = 'scores';
                this.scores = source;
            }
        }
        else if (source instanceof SelectorPickClass) {
            this.primarySource = 'selector';
            this.selector = source._toSelector();
        }
        else {
            this.primarySource = 'data';
            this.data = source;
        }
        if (options) {
            if (options.sources) {
                if (this.primarySource !== 'known' && options.sources[this.primarySource]) {
                    throw new Error('Attempted to set UUID source of the same type as the primary source in options!');
                }
                if (options.sources.scores) {
                    this.setScores(options.sources.scores === true ? undefined : options.sources.scores);
                }
                if (options.sources.data) {
                    this.setData(options.sources.data === true ? undefined : options.sources.data);
                }
                if (options.sources.selector) {
                    this.selector = options.sources.selector;
                }
            }
        }
    }
    /**
     * Converts a string UUID to an integer array
     * @param uuid A UUID in string form
     */
    stringToArray(uuid) {
        uuid = uuid.trim();
        if (uuid.includes('-')) {
            uuid = uuid
                .split('-')
                .map((g, i) => g.padStart(UUID_GROUP_SIZES[i], '0'))
                .join('');
        }
        else {
            uuid = uuid.padStart(32, '0');
        }
        UUIDData.setBigUint64(0, BigInt(`0x${uuid.substring(0, 16)}`), false);
        UUIDData.setBigUint64(8, BigInt(`0x${uuid.substring(16)}`), false);
        const array = [];
        for (let i = 0; i < 4; i++) {
            array.push(UUIDData.getInt32(i * 4, false));
        }
        return array;
    }
    /**
     * Converts an integer array to a string UUID
     * @param array Optional. An integer array. Defaults to contained UUID (known; this does not work at pack runtime)
     */
    arrayToString(array) {
        if (this.known || array) {
            if (!array) {
                array = this.known;
            }
            for (let i = 0; i < 4; i++) {
                UUIDData.setInt32(i * 4, array[i], false);
            }
            const hex = UUIDData.getBigUint64(0, false).toString(16).padStart(16, '0') + UUIDData.getBigUint64(8, false).toString(16).padStart(16, '0');
            const groups = [];
            let groupStart = 0;
            for (const groupSize of UUID_GROUP_SIZES) {
                groups.push(hex.substring(groupStart, groupStart + groupSize));
                groupStart += groupSize;
            }
            return groups.join('-');
        }
        throw new Error('Cannot convert a non-existent UUID array to a string!');
    }
    setKnown(uuid) {
        this.known = uuid;
        return this;
    }
    setScores(scores) {
        const { core } = this;
        const { Data, getTempStorage, initMCFunction, Variable, } = core.pack;
        const handleConversions = (_scores) => {
            switch (this.primarySource) {
                case 'known':
                    {
                        for (let i = 0; i < 4; i++) {
                            _scores[i].set(this.known[i]);
                        }
                    }
                    break;
                case 'data':
                    {
                        const storagePoint = this.data;
                        if (storagePoint.type !== 'storage') {
                            const dataPoint = getTempStorage('UUID');
                            dataPoint.set(storagePoint);
                        }
                        for (let i = 0; i < 4; i++) {
                            _scores[i].set(storagePoint.select([i]));
                        }
                    }
                    break;
                case 'selector':
                    {
                        const storagePoint = getTempStorage('UUID');
                        storagePoint.set(Data('entity', this.selector, 'UUID'));
                        for (let i = 0; i < 4; i++) {
                            _scores[i].set(storagePoint.select([i]));
                        }
                    }
                    break;
                default: break;
            }
        };
        // Generating & setting scores
        if (scores === undefined) {
            handleConversions([Variable(), Variable(), Variable(), Variable()]);
            // Setting existing scores
        }
        else if (Array.isArray(scores[0])) {
            // UUIDClass is being constructed in an MCFunction
            if (core.mcfunctionStack.length !== 0) {
                /* @ts-ignore */
                core.getCurrentMCFunctionOrThrow().resource.push(() => handleConversions(scores[0]));
                // UUIDClass is being constructed outside of a MCFunction
            }
            else {
                /* @ts-ignore */
                initMCFunction.push(() => handleConversions(scores[0]));
            }
        }
        else {
            this.scores = scores;
        }
        return this;
    }
    setData(data) {
        if (Array.isArray(data) || data === undefined) {
            const { Data, DataVariable, ResolveNBT, } = this.core.pack;
            // User doesn't want to set up a storage location
            if (data === undefined) {
                this.data = DataVariable(undefined, '__sandstone');
                // User has set a storage location
            }
            else {
                this.data = data[0];
            }
            switch (this.primarySource) {
                case 'known':
                    {
                        this.data.set(new NBTIntArray(this.known));
                    }
                    break;
                case 'scores':
                    {
                        ResolveNBT(this.scores.map((score) => ResolveNBTPart(score)), this.data);
                    }
                    break;
                case 'selector':
                    {
                        this.data.set(Data('entity', this.selector, 'UUID'));
                    }
                    break;
                default: break;
            }
        }
        else {
            this.data = data;
        }
        return this;
    }
    /** Selector for the entity with the UUID. If `known` is set this will go unused. */
    setSelector(selector) {
        this.selector = selector;
        return this;
    }
    /** Partial execute command executing as (not at) the entity */
    get execute() {
        const { execute } = this.core.pack.commands;
        if (this.known) {
            return execute.as(this.arrayToString());
        }
        if (this.selector) {
            return execute.as(this.selector);
        }
        // TODO: Refactor to use macro
        return execute.as;
    }
    toString() {
        if (this.known) {
            return JSON.stringify(this.known);
        }
        return this.selector?.toString() || 'Stored in scores or data storage';
    }
    /**
     * @internal
     */
    _toChatComponent() {
        if (this.known) {
            return { selector: this.arrayToString() };
        }
        if (this.selector) {
            return { selector: this.selector.toString() };
        }
        if (this.scores) {
            return [...this.scores];
        }
        return this.data;
    }
    /**
     * @internal
     */
    /* @ts-ignore */
    _toMinecraftCondition = () => (new this.core.pack.conditions.UUID(this.core, this));
    /**
     * @internal
     */
    _toSelector() {
        const selector = this.known ? this.arrayToString() : this.selector;
        if (!selector) {
            throw new Error('Cannot create a selector without a UUID or existing selector');
        }
        return selector;
    }
}
