/// <reference types="node" resolution-mode="require"/>
import * as util from 'util';
import type { NBTObject, RootNBT } from 'sandstone/arguments';
import type { MacroArgument } from 'sandstone/core';
export declare abstract class NBTClass {
    abstract [util.inspect.custom]: () => string;
    toString: () => string;
}
export declare class NBTPrimitive extends NBTClass {
    value: number;
    unit: string;
    constructor(value: number, unit: string);
    [util.inspect.custom]: () => string;
}
export declare class NBTLong extends NBTPrimitive {
    constructor(value: number);
}
export declare class NBTByte extends NBTPrimitive {
    constructor(value: number);
}
export declare class NBTShort extends NBTPrimitive {
    constructor(value: number);
}
export declare class NBTFloat extends NBTPrimitive {
    constructor(value: number);
}
export declare class NBTDouble extends NBTPrimitive {
    constructor(value: number);
}
export declare class NBTTypedArray extends NBTClass {
    values: number[];
    unit: string;
    constructor(values: number[], unit: string);
    [util.inspect.custom]: () => string;
}
export declare class NBTLongArray extends NBTTypedArray {
    constructor(values: number[]);
}
export declare class NBTByteArray extends NBTTypedArray {
    constructor(values: number[]);
}
export declare class NBTIntArray extends NBTTypedArray {
    constructor(values: number[]);
}
export declare class NotNBT extends NBTClass {
    nbt: RootNBT;
    constructor(nbt: RootNBT);
    [util.inspect.custom]: () => string;
}
export type NBTSimpleClasses = typeof NBTLong | typeof NBTShort | typeof NBTByte | typeof NBTFloat | typeof NBTDouble;
export declare class NBTInt extends NBTPrimitive {
}
export type NBTAllNumbers = typeof NBTInt | NBTSimpleClasses;
export type NBTAllArrays = typeof NBTTypedArray | typeof NBTLongArray | typeof NBTByteArray | typeof NBTIntArray;
export declare class NBTString extends NBTPrimitive {
}
export type NBTAllPrimitives = NBTAllNumbers | NBTAllArrays | typeof NBTString;
export declare class NBTAnyValue extends NBTPrimitive {
}
export type NBTAllValues = NBTAllPrimitives | NBTAnyValue;
interface NBTInterface {
    /**
     * Transform a number into a Minecraft NBT float number.
     *
     * @param floatNumber The number to transform.
     *
     * @example
     * summon(..., { Health: 20 }) // => { Health: 20 }
     *
     * summon(..., { Health: NBT.float(20) }) // => { Health: 20f }
     */
    float(floatNumber: number): NBTFloat;
    /**
     * Transform several numbers into a an array of Minecraft NBT float numbers.
     *
     * @param floatNumbers The numbers to transform.
     *
     * @example
     * summon(..., { Test: [0, 1, 2] }) // => { Test: [0, 1, 2] }
     *
     * summon(..., { Test: NBT.float([0, 1, 2]) }) // => { Test: [0f, 1f, 2f] }
     */
    float(floatNumbers: number[]): NBTFloat[];
    /**
     * Transform a number into a Minecraft NBT double number.
     *
     * @param doubleNumber The number to transform.
     *
     * @example
     * summon(..., { Health: 20 }) // => { Health: 20 }
     *
     * summon(..., { Health: NBT.double(20) }) // => { Health: 20d }
     */
    double(doubleNumber: number): NBTDouble;
    /**
     * Transform several numbers into a an array of Minecraft NBT double numbers.
     *
     * @param doubleNumbers The numbers to transform.
     *
     * @example
     * summon(..., { Test: [0, 1, 2] }) // => { Test: [0, 1, 2] }
     *
     * summon(..., { Test: NBT.double([0, 1, 2]) }) // => { Test: [0d, 1d, 2d] }
     */
    double(doubleNumbers: number[]): NBTDouble[];
    /**
     * Transform a number into a Minecraft NBT byte number.
     *
     * @param byteNumber The number to transform.
     *
     * @example
     * summon(..., { Health: 20 }) // => { Health: 20 }
     *
     * summon(..., { Health: NBT.byte(20) }) // => { Health: 20b }
     */
    byte(byteNumber: number): NBTByte;
    /**
     * Transform several numbers into a an array of Minecraft NBT byte numbers.
     *
     * @param byteNumbers The numbers to transform.
     *
     * @example
     * summon(..., { Test: [0, 1, 2] }) // => { Test: [0, 1, 2] }
     *
     * summon(..., { Test: NBT.byte([0, 1, 2]) }) // => { Test: [0b, 1b, 2b] }
     */
    byte(byteNumbers: number[]): NBTByte[];
    /**
     * Transform a number into a Minecraft NBT short number.
     *
     * @param shortNumber The number to transform.
     *
     * @example
     * summon(..., { Health: 20 }) // => { Health: 20 }
     *
     * summon(..., { Health: NBT.short(20) }) // => { Health: 20b }
     */
    short(shortNumber: number): NBTShort;
    /**
     * Transform several numbers into a an array of Minecraft NBT short numbers.
     *
     * @param shortNumbers The numbers to transform.
     *
     * @example
     * summon(..., { Test: [0, 1, 2] }) // => { Test: [0, 1, 2] }
     *
     * summon(..., { Test: NBT.short([0, 1, 2]) }) // => { Test: [0s, 1s, 2s] }
     */
    short(shortNumbers: number[]): NBTShort[];
    /**
     * Transform a number into a Minecraft NBT long number.
     *
     * @param longNumber The number to transform.
     *
     * @example
     * summon(..., { Health: 20 }) // => { Health: 20 }
     *
     * summon(..., { Health: NBT.long(20) }) // => { Health: 20l }
     */
    long(longNumber: number): NBTLong;
    /**
     * Transform several numbers into a an array of Minecraft NBT long numbers.
     *
     * @param longNumbers The numbers to transform.
     *
     * @example
     * summon(..., { Test: [0, 1, 2] }) // => { Test: [0, 1, 2] }
     *
     * summon(..., { Test: NBT.long([0, 1, 2]) }) // => { Test: [0l, 1l, 2l] }
     */
    long(longNumbers: number[]): NBTLong[];
    /**
     * Transforms an array into an Integer array.
     *
     * @param intNumbers The numbers to transform.
     *
     * @example
     * summon(..., { Test: [0, 1, 2] }) // => { Test: [0, 1, 2] }
     *
     * summon(..., { Test: NBT.intArray([0, 1, 2]) }) // => { Test: [I; 0, 1, 2] }
     */
    intArray(intNumbers: number[]): NBTIntArray;
    /**
     * Transforms an array into a Byte array.
     *
     * @param intNumbers The numbers to transform.
     *
     * @example
     * summon(..., { Test: [0, 1, 2] }) // => { Test: [0, 1, 2] }
     *
     * summon(..., { Test: NBT.byteArray([0, 1, 2]) }) // => { Test: [B; 0, 1, 2] }
     */
    byteArray(intNumbers: number[]): NBTByteArray;
    /**
     * Transforms an array into a Long array.
     *
     * @param longNumbers The numbers to transform.
     *
     * @example
     * summon(..., { Test: [0, 1, 2] }) // => { Test: [0, 1, 2] }
     *
     * summon(..., { Test: NBT.longArray([0, 1, 2]) }) // => { Test: [L; 0, 1, 2] }
     */
    longArray(longNumbers: number[]): NBTLongArray;
    /**
     * Used to check for the absence of a NBT in a selector.
     *
     * @param nbt The NBT to check the absence of.
     *
     * @example
     *
     * // Check for entities that are NOT new babies.
     * >>> Selector('@e', { nbt: NBT.not({ Age: -32768 }) }).toString()
     * @e[nbt=!{ Age: -32768 }]
     */
    not: (nbt: RootNBT) => NotNBT;
    /**
     * Return a stringified version of a NBT.
     *
     * @param nbt The NBT to stringify.
     *
     * @example
     *
     * >>> NBT.stringify({ Invisible: NBT.byte(1) })
     * '{ Invisible: 1b }'
     *
     * >>> NBT.stringify({ Rotation: NBT.float([ 90, 0 ]) })
     * '{ Rotation: [90f, 0f] }'
     *
     * >>> NBT.stringify({ Test: NBT.longArray([0, 1, 2]) })
     * '{ Test: [L; 0, 1, 2] }'
     */
    stringify: (nbt: NBTObject) => string;
    /**
     * Parses the given NBT.
     *
     * @param nbt The NBT to parse.
     *
     * @example
     *
     * >>> NBT.parse(`{ Invisible: NBT.byte(1) }`)
     * { Invisible: 1b }
     *
     * >>> const inv = NBT.parse(`[{ Slot: 1b, Count: 1, id: 'minecraft:dirt' }]`)
     * [{ Slot: 1b, Count: 1, id: 'minecraft:dirt' }]
     * >>> inv.length
     * 1
     *
     * >>> NBT.parse('{ Test: [L; 0, 1, 2] }')
     * { Test: [L; 0, 1, 2] }
     */
    parse: (nbt: string) => NBTObject;
    /**
     * Parses the given NBT.
     *   *
     * @example
     *
     * >>> NBT`{ Invisible: NBT.byte(1) }`
     * { Invisible: 1b }
     *
     * >>> const inv = NBT`[{ Slot: 1b, Count: 1, id: 'minecraft:dirt' }]`
     * [{ Slot: 1b, Count: 1, id: 'minecraft:dirt' }]
     * >>> inv.length
     * 1
     *
     * >>> NBT`{ Test: [L; 0, 1, 2] }`
     * { Test: [L; 0, 1, 2] }
     */
    (nbts: TemplateStringsArray, ...args: unknown[]): NBTObject;
}
export declare const NBT: NBTInterface;
export declare const nbtStringifier: (nbt: NBTObject | MacroArgument) => string;
export {};
