import { makeCallable } from 'sandstone/utils';
import * as util from 'util';
import { parseNBT } from './parser.js';
export class NBTClass {
    toString = this[util.inspect.custom];
}
export class NBTPrimitive extends NBTClass {
    value;
    unit;
    constructor(value, unit) {
        super();
        this.value = value;
        this.unit = unit;
    }
    [util.inspect.custom] = () => `${this.value}${this.unit}`;
}
export class NBTLong extends NBTPrimitive {
    constructor(value) { super(value, 'l'); }
}
export class NBTByte extends NBTPrimitive {
    constructor(value) { super(value, 'b'); }
}
export class NBTShort extends NBTPrimitive {
    constructor(value) { super(value, 's'); }
}
export class NBTFloat extends NBTPrimitive {
    constructor(value) { super(value, 'f'); }
}
export class NBTDouble extends NBTPrimitive {
    constructor(value) { super(value, 'd'); }
}
export class NBTTypedArray extends NBTClass {
    values;
    unit;
    constructor(values, unit) {
        super();
        this.values = values;
        this.unit = unit;
    }
    [util.inspect.custom] = () => `[${this.unit};${this.values.length === 0 ? '' : this.values.join(',')}]`;
}
export class NBTLongArray extends NBTTypedArray {
    constructor(values) { super(values, 'L'); }
}
export class NBTByteArray extends NBTTypedArray {
    constructor(values) { super(values, 'B'); }
}
export class NBTIntArray extends NBTTypedArray {
    constructor(values) { super(values, 'I'); }
}
export class NotNBT extends NBTClass {
    nbt;
    constructor(nbt) {
        super();
        this.nbt = nbt;
    }
    [util.inspect.custom] = () => `!${nbtStringifier(this.nbt)}`;
}
export class NBTInt extends NBTPrimitive {
}
export class NBTString extends NBTPrimitive {
}
// Type Hack, not actually a primitive
export class NBTAnyValue extends NBTPrimitive {
}
function customNumber(num, _class) {
    if (Array.isArray(num)) {
        return num.map((n) => new _class(n));
    }
    return new _class(num);
}
function dynamicNBT(template, ...args) {
    const mixedArgs = template.flatMap((s, i) => [s, args[i]]).slice(0, -1); // Remove the last argument, which will always be undefined
    const result = mixedArgs.map((element) => (element instanceof NBTClass ? nbtStringifier(element) : element.toString())).join('');
    return parseNBT(NBT, result);
}
// TODO: Look into the weird typing on this
export const NBT = makeCallable({
    float: (num) => customNumber(num, NBTFloat),
    double: (num) => customNumber(num, NBTDouble),
    byte: (num) => customNumber(num, NBTByte),
    short: (num) => customNumber(num, NBTShort),
    long: (num) => customNumber(num, NBTLong),
    intArray: (numbers) => new NBTIntArray(numbers),
    longArray: (numbers) => new NBTLongArray(numbers),
    byteArray: (numbers) => new NBTByteArray(numbers),
    not: (nbt) => new NotNBT(nbt),
    stringify: (nbt) => nbtStringifier(nbt),
    parse: (nbt) => parseNBT(NBT, nbt),
}, dynamicNBT);
export const nbtStringifier = (nbt) => {
    if (typeof nbt === 'number') {
        // We have a number
        return nbt.toString();
    }
    if (typeof nbt === 'boolean') {
        // We have a boolean
        return nbt.toString();
    }
    if (typeof nbt === 'string') {
        // We have a string
        /*
         * Sometimes, when we have both a " and a ' in a string,
         * util.inspect will end up creating a template string, invalid for Minecraft.
         */
        const inspectedStr = util.inspect(nbt, {
            breakLength: +Infinity,
            compact: true,
            maxStringLength: +Infinity,
            depth: +Infinity,
        });
        if (inspectedStr[0] === '`') {
            return JSON.stringify(nbt);
        }
        return inspectedStr;
    }
    if (Array.isArray(nbt)) {
        // We have an array
        const itemsStr = nbt.map(nbtStringifier).join(',');
        return `[${itemsStr}]`;
    }
    // We have an object
    if (nbt instanceof NBTClass) {
        // It's actually a "Minecraft Primitive", like 1b, and not an object
        return nbt[util.inspect.custom]();
    }
    if (typeof nbt === 'object' && nbt.toMacro) {
        /* @ts-ignore */
        return nbt.toMacro();
    }
    // It's a real object.
    const objectStr = Object.entries(nbt).map(([key, value]) => `${key}:${nbtStringifier(value)}`).join(',');
    return `{${objectStr}}`;
};
