import { isMacroArgument } from '../core/Macro.js';
import { VectorClass } from './Coordinates.js';
// PARSERS
export function arrayToArgsParser(args) {
    if (Array.isArray(args) && args.length === 3) {
        return new VectorClass(args.map((arg) => arg.toString()));
    }
    return args;
}
function isRawCoordinates(arg) {
    return Array.isArray(arg) && (arg.length === 3 || arg.length === 2) && arg.every((c) => typeof c === 'string');
}
function isRawRotation(arg) {
    return Array.isArray(arg) && arg.length === 2 && arg.every((c) => typeof c === 'string');
}
export function coordinatesParser(coordinates) {
    return isRawCoordinates(coordinates) ? new VectorClass(coordinates) : coordinates;
}
export function rotationParser(rotation) {
    return isRawRotation(rotation) ? new VectorClass(rotation) : rotation;
}
// Sanitize score values. null => '', Infinity => '', any number => itself
export const sanitizeValue = (core, value) => {
    if (value === undefined || value === null) {
        return '';
    }
    if (typeof value === 'string') {
        return value;
    }
    if (isMacroArgument(core, value)) {
        return value.toMacro();
    }
    if (Number.isFinite(value)) {
        return value.toString();
    }
    // Value is Infinity or -Infinity
    return '';
};
// Returns the string representation of a range. [0, null] => '0..', [-Infinity, 5] => '..5', 8 => '8'
export const rangeParser = (core, range) => {
    if (Array.isArray(range)) {
        return `${sanitizeValue(core, range[0])}..${sanitizeValue(core, range[1])}`;
    }
    return range.toString();
};
export const structureRotationParser = (rotation) => {
    if (!rotation) {
        return 'none';
    }
    if (typeof rotation === 'object') {
        return rotation;
    }
    const numToLiteral = (angle) => {
        switch (angle) {
            case 0: return 'none';
            case 90: return 'clockwise_90';
            case 180: return '180';
            case 270: return 'counterclockwise_90';
            case -90: return 'counterclockwise_90';
            case -180: return '180';
            case -270: return 'clockwise_90';
            default: {
                if (!Number.isInteger(angle / 90)) {
                    throw new Error('Structure rotation must be in increments of 90!');
                }
                // reduce the angle
                angle %= 360;
                // force it to be the positive remainder, so that 0 <= angle < 360
                angle = (angle + 360) % 360;
                // force into the minimum absolute value residue class, so that -180 < angle <= 180
                if (angle > 180)
                    angle -= 360;
                return numToLiteral(angle);
            }
        }
    };
    if (typeof rotation === 'number') {
        return numToLiteral(rotation);
    }
    if (rotation === 'clockwise_90' || rotation === 'counterclockwise_90') {
        return rotation;
    }
    return numToLiteral(Number(rotation));
};
export const structureMirrorParser = (mirror) => {
    if (typeof mirror === 'object') {
        return mirror.toMacro();
    }
    if (typeof mirror === 'string') {
        const lastCharacter = mirror.slice(0, -1);
        if (lastCharacter === 'x') {
            return 'left_right';
        }
        if (lastCharacter === 'z') {
            return 'front_back';
        }
        return mirror;
    }
    if (mirror) {
        return 'left_right';
    }
    return 'none';
};
export function targetParser(target) {
    if (typeof target === 'string') {
        return target;
    }
    if (target._toSelector) {
        return target._toSelector();
    }
    if (target.toMacro) {
        return target;
    }
    return target.toString();
}
